#!/usr/bin/env python3

from datetime import datetime
from threading import Thread, Event
from collections import defaultdict, namedtuple
from urllib.parse import urlsplit, urlparse
from enum import Enum, IntEnum

from .message import LogLevel, JobIOMessage, ControlMessage, SessionEndMessage, Message, MessageType
from .pdu import PDUState
from .message import JobStatus
from .job import Job
from .logger import logger
from .minioclient import MinioClient, MinIOPolicyStatement, generate_policy
from . import config

import traceback
import requests
import tempfile
import secrets
import random
import select
import shutil
import socket
import json
import time


# Constants
CONSOLE_DRAINING_DELAY = 1


class MachineState(Enum):
    WAIT_FOR_CONFIG = 0
    IDLE = 1
    TRAINING = 2
    QUEUED = 3
    RUNNING = 4


def str_to_int(string, default):
    try:
        return int(string)
    except Exception:
        return default


class JobConsoleState(IntEnum):
    CREATED = 0
    ACTIVE = 1
    DUT_DONE = 2
    TEAR_DOWN = 3
    OVER = 4


class JobConsole(Thread):
    def __init__(self, machine_id, client_endpoint, console_patterns,
                 client_version=None, log_level=LogLevel.INFO):
        super().__init__(name='ConsoleThread')

        self.machine_id = machine_id

        self.client_endpoint = client_endpoint
        self.console_patterns = console_patterns
        self.client_version = client_version
        self.log_level = log_level

        # Sockets
        self.client_sock = None
        self.salad_sock = self.connect_to_salad()

        # Job-long state
        self._state = JobConsoleState.CREATED
        self.start_time = None
        self.line_buffer = bytearray()
        self._user_session_state = dict()

        self.reset_per_boot_state()

    @property
    def salad_url(self):
        return f"{config.SALAD_URL}/api/v1/machine/{self.machine_id}"

    def connect_to_salad(self):
        parsed_url = urlsplit(config.SALAD_URL)

        r = requests.get(self.salad_url)
        r.raise_for_status()

        machine = r.json()
        port = machine.get("tcp_port")

        return socket.create_connection((parsed_url.hostname, port))

    def reset_per_boot_state(self):
        self.last_activity_from_machine = None
        self.last_activity_from_client = None

        self.console_patterns.reset_per_boot_state()
        self.needs_reboot = self.console_patterns.needs_reboot

    def close_salad(self):
        try:
            self.salad_sock.shutdown(socket.SHUT_RDWR)
            self.salad_sock.close()
        except OSError:
            pass

    def close_client(self):
        if self.client_version:
            try:
                self.client_sock.shutdown(socket.SHUT_RDWR)
                self.client_sock.close()
            except OSError:
                pass

    def close(self):
        self.set_state(JobConsoleState.OVER)

    @property
    def state(self):
        if self._state == JobConsoleState.ACTIVE:
            return self._state if self.is_alive() else JobConsoleState.OVER

        return self._state

    def set_state(self, state, **kwargs):
        prev_state = self._state
        if state < prev_state:
            raise ValueError("The state can only move forward")
        elif state == prev_state:
            return
        else:
            self._state = state

        self.log(f"Job console state changed from {prev_state.name} -> {state.name}\n")

        if state == JobConsoleState.ACTIVE:
            self.start_time = datetime.now()

        elif state == JobConsoleState.DUT_DONE:
            # Skip the entire tear-down if we do not have a client
            if not self.client_version:
                self.set_state(JobConsoleState.OVER)

        elif state == JobConsoleState.TEAR_DOWN:
            # Kill the connection to SALAD
            self.close_salad()

            # Notify the client
            if self.client_version:
                if self.client_version == 0:
                    self.log(f"<-- End of the session: {self.console_patterns.job_status} -->\n")
                elif self.client_version == 1:
                    try:
                        status = JobStatus.from_str(self.console_patterns.job_status)
                        SessionEndMessage.create(job_bucket=kwargs.get('job_bucket'),
                                                 status=status).send(self.client_sock)
                    except (ConnectionResetError, BrokenPipeError, OSError):
                        traceback.print_exc()
                try:
                    self.client_sock.shutdown(socket.SHUT_WR)
                except (ConnectionResetError, BrokenPipeError, OSError):
                    pass

        elif state == JobConsoleState.OVER:
            # Make sure the connections to SALAD and the client are killed
            self.close_salad()
            self.close_client()

    def start(self):
        if self.client_version:
            logger.info(f"Connecting to the client endpoint {self.client_endpoint}")
            self.client_sock = socket.create_connection(self.client_endpoint)
        super().start()

    def match_console_patterns(self, buf):
        patterns_matched = set()

        # Process the buffer, line by line
        to_process = self.line_buffer + buf
        cur = 0
        while True:
            idx = to_process.find(b'\n', cur)
            if idx > 0:
                line = to_process[cur:idx+1]
                logger.info(f"{self.machine_id} -> {bytes(line)}")
                patterns_matched |= self.console_patterns.process_line(line)
                cur = idx + 1
            else:
                break
        self.line_buffer = to_process[cur:]

        # Tell the user what happened
        if len(patterns_matched) > 0:
            self.log(f"Matched the following patterns: {', '.join(patterns_matched)}\n")

        # Check if the state changed
        self.needs_reboot = self.console_patterns.needs_reboot

    def log(self, msg, log_level=LogLevel.INFO):
        # Ignore messages with a log level lower than the minimum set
        if log_level < self.log_level:
            return

        if self.start_time is not None:
            relative_time = (datetime.now() - self.start_time).total_seconds()
        else:
            relative_time = 0.0

        log_msg = f"+{relative_time:.3f}s: {msg}"
        logger.info(log_msg.rstrip("\r\n"))

        if self.client_version:
            try:
                if self.client_version == 0:
                    self.client_sock.send(log_msg.encode())
                elif self.client_version == 1:
                    ControlMessage.create(log_msg, severity=log_level).send(self.client_sock)
            except OSError:
                pass

    def stop(self):
        self.set_state(JobConsoleState.OVER)
        self.join()

    def run(self):
        self.set_state(JobConsoleState.ACTIVE)

        while self.state < JobConsoleState.OVER:
            fds = []
            if self.state < JobConsoleState.TEAR_DOWN:
                fds.extend([self.salad_sock.fileno()])
            if self.client_version:
                fds.extend([self.client_sock.fileno()])

            # Make sure all the FDs are valid, or exit!
            if any([fd < 0 for fd in fds]):
                self.log("Found a negative fd, aborting!")
                self.close()

            rlist, _, _ = select.select(fds, [], [], 1.0)

            for fd in rlist:
                try:
                    if fd == self.salad_sock.fileno():
                        # DUT's stdout/err: Salad -> Client
                        buf = self.salad_sock.recv(8192)
                        if len(buf) == 0:
                            self.set_state(JobConsoleState.DUT_DONE)

                        # Match the console patterns
                        try:
                            self.match_console_patterns(buf)
                        except Exception:
                            self.log(traceback.format_exc())

                        # Update the last console activity if we already had activity,
                        # or when we get the first newline character as serial
                        # consoles may sometimes send unwanted characters at power up
                        if self.last_activity_from_machine is not None or b'\n' in buf:
                            self.last_activity_from_machine = datetime.now()

                        # Forward to the client
                        if self.client_version:
                            if self.client_version == 0:
                                self.client_sock.send(buf)
                            elif self.client_version == 1:
                                JobIOMessage.create(buf).send(self.client_sock)

                        # The message got forwarded, close the session if it ended
                        if self.console_patterns.session_has_ended:
                            self.set_state(JobConsoleState.DUT_DONE)

                    elif fd == self.client_sock.fileno():
                        # DUT's stdin: Client -> Salad
                        if self.client_version == 0:
                            buf = self.client_sock.recv(8192)
                            if len(buf) == 0:
                                self.close()

                            # Forward to the salad
                            self.salad_sock.send(buf)
                        elif self.client_version == 1:
                            try:
                                msg = Message.next_message(self.client_sock)
                                if msg.msg_type == MessageType.JOB_IO:
                                    self.salad_sock.send(msg.buffer)
                            except EOFError:
                                # Do not warn when we are expecting the client to close its socket
                                if self.state < JobConsoleState.TEAR_DOWN:
                                    self.log(traceback.format_exc())

                                self.log("The client closed its connection")

                                # Clean up everything on our side
                                self.close()

                        self.last_activity_from_client = datetime.now()
                except (ConnectionResetError, BrokenPipeError, OSError):
                    self.log(traceback.format_exc())
                    self.close()
                except Exception:
                    logger.error(traceback.format_exc())


class SergentHartman:
    def __init__(self, machine, boot_loop_counts=None, qualifying_rate=None):
        super().__init__()

        if boot_loop_counts is None:
            boot_loop_counts = int(config.SERGENT_HARTMAN_BOOT_COUNT)

        if qualifying_rate is None:
            qualifying_rate = int(config.SERGENT_HARTMAN_QUALIFYING_BOOT_COUNT)

        self.machine = machine
        self.boot_loop_counts = boot_loop_counts
        self.qualifying_rate = qualifying_rate

        self.reset()

    @property
    def is_machine_registered(self):
        return self.cur_loop > 0

    def reset(self):
        self.is_active = False
        self.cur_loop = 0
        self.statuses = defaultdict(int)

    def next_task(self):
        mid = self.machine.id

        if not self.is_active:
            # Start by forcing the machine to register itself to make sure the
            # its configuration is up to date (especially the serial console
            # port). Loop until it succeeds!
            self.reset()

            logger.info("SergentHartman/%s - Try registering the machine", mid)

            self.is_active = True

            return Job.from_path(config.EXECUTOR_REGISTRATION_JOB, self.machine)
        else:
            # Check that we got the expected amount of reports
            if self.cur_loop != sum(self.statuses.values()):
                raise ValueError("The previous next_task() call was not followed by a call to report()")

            # The registration went well, let's start the boot loop!
            self.cur_loop += 1

            statuses_str = [f"{status.name}: {values}" for status, values in self.statuses.items()]
            logger.info("SergentHartman/%s - loop %s/%s - statuses %s: "
                        "Execute one more round!",
                        mid,
                        self.cur_loop,
                        self.boot_loop_counts,
                        statuses_str)

            return Job.from_path(config.EXECUTOR_BOOTLOOP_JOB, self.machine)

    def report(self, job_status):
        mid = self.machine.id

        if self.cur_loop == 0:
            if job_status != JobStatus.PASS:
                delay = int(config.SERGENT_HARTMAN_REGISTRATION_RETRIAL_DELAY)
                logger.warning((f"SergentHartman/{mid} - Registration failed with status {job_status.name}. "
                                f"Retrying in {delay} second(s)"))
                self.reset()
                return delay
            else:
                logger.info(f"SergentHartman/{mid} - Registration succeeded, moving on to the boot loop")
        else:
            # We are in the boot loop
            self.statuses[job_status] += 1

            if self.cur_loop == self.boot_loop_counts:
                self.is_active = False

                # Update MaRS
                ready_for_service = self.statuses[JobStatus.PASS] >= self.qualifying_rate
                self.machine.ready_for_service = ready_for_service

        return 0

    @property
    def is_available(self):
        return config.EXECUTOR_REGISTRATION_JOB or config.EXECUTOR_BOOTLOOP_JOB


class JobBucket:
    Credentials = namedtuple('Credentials', ['username', 'password', 'policy_name'])

    def __init__(self, minio, bucket_name, initial_state_tarball_file=None):
        self.minio = minio
        self.name = bucket_name

        self._credentials = dict()

        if initial_state_tarball_file:
            self.initial_state_tarball_file = tempfile.NamedTemporaryFile("w+b")
            shutil.copyfileobj(initial_state_tarball_file, self.initial_state_tarball_file)
            self.initial_state_tarball_file.seek(0)
        else:
            self.initial_state_tarball_file = None

        self.minio.make_bucket(bucket_name)

    def remove(self):
        self.minio.remove_bucket(self.name)

        for credentials in self._credentials.values():
            self.minio.remove_user_policy(credentials.policy_name, credentials.username)
            self.minio.remove_user(credentials.username)

    def __del__(self):
        try:
            self.remove()
        except Exception:
            traceback.print_exc()

    def credentials(self, role):
        return self._credentials.get(role)

    def create_owner_credentials(self, role, user_name=None, password=None,
                                 groups=None, whitelisted_ips=None):
        if user_name is None:
            user_name = f"{self.name}-{role}"

        if password is None:
            password = secrets.token_hex(16)

        if groups is None:
            groups = []

        if whitelisted_ips is None:
            whitelisted_ips = []

        policy_name = f"policy_{user_name}"

        self.minio.add_user(user_name, password)

        policy_statements = [
            MinIOPolicyStatement(buckets=[self.name], source_ips=whitelisted_ips)
        ]
        if len(whitelisted_ips) > 0:
            restrict_to_whitelisted_ips = MinIOPolicyStatement(allow=False, not_source_ips=whitelisted_ips)
            policy_statements.append(restrict_to_whitelisted_ips)
        policy = json.dumps(generate_policy(policy_statements))
        logger.debug(f"Applying the MinIO policy: {policy}")

        try:
            self.minio.apply_user_policy(policy_name, user_name, policy_statements)
        except Exception as e:
            self.minio.remove_user(user_name)
            raise e from None

        # Add the user to the wanted list of groups
        for group_name in groups:
            self.minio.add_user_to_group(user_name, group_name)

        credentials = self.Credentials(user_name, password, policy_name)
        self._credentials[role] = credentials

        return credentials

    def setup(self):
        if self.initial_state_tarball_file:
            self.minio.extract_archive(self.initial_state_tarball_file, self.name)

    def access_url(self, role=None):
        endpoint = urlparse(self.minio.url)

        role_creds = self.credentials(role)
        if role_creds:
            credentials = f"{role_creds[0]}:{role_creds[1]}@"
        else:
            credentials = ""
        return f'{endpoint.scheme}://{credentials}{endpoint.netloc}'

    @classmethod
    def from_job_request(cls, minio, request, machine):
        # Generate a job id
        bucket_name = MinioClient.create_valid_bucket_name(f"job-{machine.id}-{request.job_id}")

        try:
            # BUG: It seems like this is not a reliable way to detect re-use of existing buckets...
            return cls(minio, bucket_name=bucket_name,
                       initial_state_tarball_file=request.job_bucket_initial_state_tarball_file)
        except ValueError:
            # The bucket already exists, let's try to make it more unique!
            now = int(datetime.utcnow().timestamp())
            rand_int = random.randrange(10e6)
            return cls(minio, bucket_name=f"{bucket_name}-{now}-{rand_int}",
                       initial_state_tarball_file=request.job_bucket_initial_state_tarball_file)


class Executor(Thread):
    def __init__(self, machine):
        super().__init__(name=f'ExecutorThread-{machine.id}')

        self.machine = machine

        self.state = MachineState.WAIT_FOR_CONFIG
        self.minio = MinioClient()

        # Training / Qualifying process
        self.sergent_hartman = SergentHartman(machine)

        # Outside -> Inside communication
        self.job_ready = Event()
        self.job_config = None
        self.job_console = None
        self.job_bucket = None

        # Remote artifacts (typically over HTTPS) are stored in our
        # local minio instance which is exposed over HTTP to the
        # private LAN. This makes such artifacts amenable to PXE
        # booting, for which HTTPS clients are not available.  Less
        # critically, it makes access easier for the boards in our
        # private LAN, for which HTTPS offers no advantage.
        self.remote_url_to_local_cache_mapping = {}

        # Start the background thread that will manage the machine
        self.stop_event = Event()
        self.start()

    def start_job(self, job_request):
        if self.state != MachineState.IDLE:
            raise ValueError(f"The machine isn't idle: Current state is {self.state.name}")

        # Prepare the job bucket
        self.job_bucket = JobBucket.from_job_request(self.minio, job_request, self.machine)
        if self.job_bucket:
            self.job_bucket.create_owner_credentials("dut", groups=job_request.minio_groups,
                                                     whitelisted_ips=[f'{self.machine.ip_address}/32'])

        # Bit nasty to render twice, but better than duplicating
        # template render in the various call-sites within
        # executor. Rendering it up front reduces the chances for
        # mistakes. (Meta-point: using an HTTP query to specify the
        # "target" could avoid this duplication of work, and might
        # actually make more sense)
        job = Job.render_with_resources(job_request.raw_job, self.machine, self.job_bucket)
        logger.debug("rendered job:\n%s", job)

        self.state = MachineState.QUEUED
        self.job_request = job_request
        self.job_config = job
        self.job_console = JobConsole(self.machine.id,
                                      client_endpoint=job_request.callback_endpoint,
                                      console_patterns=self.job_config.console_patterns,
                                      client_version=job_request.version)
        self.job_ready.set()

    def log(self, msg, log_level=LogLevel.INFO):
        if self.job_console is not None:
            self.job_console.log(msg, log_level=log_level)

    def _cache_remote_artifact(self, artifact_name, start_url, continue_url):
        artifact_prefix = f"{artifact_name}-{self.machine.id}"

        # Assume the remote artifacts already exist locally
        self.remote_url_to_local_cache_mapping[start_url] = start_url
        self.remote_url_to_local_cache_mapping[continue_url] = continue_url

        def cache_it(url, suffix):
            if self.minio.is_local_url(url):
                logger.debug(f"Ignore caching {url} as it is already hosted by our minio cache")
                return
            self.remote_url_to_local_cache_mapping[url] = f"http://10.42.0.1:9000/boot/{artifact_prefix}-{suffix}"
            self.log(f'Caching {url} into minio...\n')
            self.minio.save_boot_artifact(start_url, f"{artifact_prefix}-start")

        cache_it(start_url, 'start')
        if start_url != continue_url:
            cache_it(continue_url, 'continue')

    def _cache_remote_artifacts(self):
        deploy_strt = self.job_config.deployment_start
        deploy_cnt = self.job_config.deployment_start

        logger.info("Caching the kernel...")
        self._cache_remote_artifact("kernel", deploy_strt.kernel_url,
                                    deploy_cnt.kernel_url)

        logger.info("Caching the initramfs...")
        self._cache_remote_artifact("initramfs", deploy_strt.initramfs_url,
                                    deploy_cnt.initramfs_url)

        if self.job_bucket:
            logger.info("Initializing the job bucket with the client's data")
            self.job_bucket.setup()

    def run(self):
        def session_init():
            # Reset the state
            self.job_config = None
            self.job_console = None
            self.machine.boots.remove_pxelinux_config(self.machine.mac_address)

            # Pick a job
            if self.sergent_hartman.is_available and not self.machine.ready_for_service:
                self.state = MachineState.TRAINING

                self.job_config = self.sergent_hartman.next_task()
                self.job_console = JobConsole(self.machine.id,
                                              client_endpoint=None,
                                              client_version=None,
                                              console_patterns=self.job_config.console_patterns)
            else:
                self.sergent_hartman.reset()

                # Wait for a job to be set
                self.state = MachineState.IDLE
                if not self.job_ready.wait(1):
                    return False
                self.job_ready.clear()

                self.state = MachineState.RUNNING

            # Cut the power to the machine, we do not need it
            self.machine.pdu_port.set(PDUState.OFF)

            # Mark the start time to now()
            self.job_start_time = datetime.now()

            # Connect to the client's endpoint, to relay the serial console
            self.job_console.start()

            return True

        def session_end():
            cooldown_delay_s = 0

            if self.sergent_hartman.is_active and self.job_config is not None:
                status = JobStatus.from_str(self.job_config.console_patterns.job_status)
                cooldown_delay_s = int(self.sergent_hartman.report(status))

            self.job_config = None

            # Signal to the job that we reached the end of the execution
            if self.job_console is not None:
                self.job_console.close()
                self.job_console = None
                self.machine.boots.remove_pxelinux_config(self.machine.mac_address)
                if self.job_bucket:
                    del self.job_bucket

            # Interruptible sleep
            for i in range(cooldown_delay_s):
                if self.stop_event.is_set():
                    return
                time.sleep(1)

        def log_exception():
            logger.debug("Exception caught:\n%s", traceback.format_exc())
            self.log(f"An exception got caught: {traceback.format_exc()}\n", LogLevel.ERROR)
            # If exceptions start firing, throttle the parent loop, since it's
            # very heavy spam if left to run at full speed.
            time.sleep(2)

        def execute_job():
            # Start the overall timeout
            timeouts = self.job_config.timeouts
            timeouts.overall.start()

            # Download the kernel/initramfs
            self.log("Setup the infrastructure\n")
            timeouts.infra_setup.start()
            self._cache_remote_artifacts()
            self.log(f"Completed setup of the infrastructure, after {timeouts.infra_setup.active_for} s\n")
            timeouts.infra_setup.stop()

            # Keep on resuming until success, timeouts' retry limits is hit, or the entire executor is going down
            deployment = self.job_config.deployment_start
            while (not self.stop_event.is_set() and not timeouts.overall.has_expired and
                   self.job_console.state < JobConsoleState.DUT_DONE):
                self.job_console.reset_per_boot_state()

                # Make sure the machine shuts down
                self.machine.pdu_port.set(PDUState.OFF)

                # Set up the deployment
                self.log("Setting up the boot configuration\n")
                self.machine.boots.write_pxelinux_config(
                    mac_addr=self.machine.id,
                    kernel_path=self.remote_url_to_local_cache_mapping.get(deployment.kernel_url),
                    cmdline=deployment.kernel_cmdline,
                    initrd_path=self.remote_url_to_local_cache_mapping.get(deployment.initramfs_url))

                self.log(f"Power up the machine, enforcing {self.machine.pdu_port.min_off_time} seconds of down time\n")
                self.machine.pdu_port.set(PDUState.ON)

                # Start the boot, and enable the timeouts!
                self.log("Boot the machine\n")
                timeouts.boot_cycle.start()
                timeouts.first_console_activity.start()
                timeouts.console_activity.stop()

                while (self.job_console.state < JobConsoleState.DUT_DONE and
                       not self.job_console.needs_reboot and
                       not self.stop_event.is_set() and not timeouts.has_expired):
                    # Update the activity timeouts, based on when was the
                    # last time we sent it data
                    if self.job_console.last_activity_from_machine is not None:
                        timeouts.first_console_activity.stop()
                        timeouts.console_activity.reset(when=self.job_console.last_activity_from_machine)

                    # Wait a little bit before checking again
                    time.sleep(0.1)

                # Cut the power
                self.machine.pdu_port.set(PDUState.OFF)

                # Increase the retry count of the timeouts that expired, and
                # abort the job if we exceeded their limits.
                abort = False
                for timeout in timeouts.expired_list:
                    retry = timeout.retry()
                    decision = "Try again!" if retry else "Abort!"
                    self.log(f"Hit the timeout {timeout} --> {decision}\n", LogLevel.ERROR)
                    abort = abort or not retry

                # Check if the DUT asked us to reboot
                if self.job_console.needs_reboot:
                    retry = timeouts.boot_cycle.retry()
                    retries_str = f"{timeouts.boot_cycle.retried}/{timeouts.boot_cycle.retries}"
                    dec = f"Boot cycle {retries_str}, go ahead!" if retry else "Exceeded boot loop count, aborting!"
                    self.log(f"The DUT asked us to reboot: {dec}\n", LogLevel.WARN)
                    abort = abort or not retry

                if abort:
                    return

                # Stop all the timeouts, except the overall
                timeouts.first_console_activity.stop()
                timeouts.console_activity.stop()
                timeouts.boot_cycle.stop()

                # We went through one boot cycle, update the
                deployment = self.job_config.deployment_continue

            # We either reached the end of the job, or the client got disconnected
            if self.job_console.state == JobConsoleState.DUT_DONE:
                # Tearing down the job
                self.log("The job has finished executing, starting tearing down\n")
                timeouts.infra_teardown.start()

                # Delay to make sure messages are read before the end of the job
                time.sleep(CONSOLE_DRAINING_DELAY)

                # Start the tear down, which will create and send the credentials
                # for the job bucket to the client
                self.log("Creating credentials to the job bucket for the client\n")
                self.job_console.set_state(JobConsoleState.TEAR_DOWN, job_bucket=self.job_bucket)

                # Wait for the client to close the connection
                self.log("Waiting for the client to download the job bucket\n")
                while (self.job_console.state < JobConsoleState.OVER and
                       not self.stop_event.is_set() and
                       not timeouts.infra_teardown.has_expired):
                    # Wait a little bit before checking again
                    time.sleep(0.1)

                self.log(f"Completed the tear down procedure in {timeouts.infra_teardown.active_for} s\n")
                timeouts.infra_teardown.stop()
            else:
                self.log("The job is over, skipping sharing the job bucket with the client")

            # We are done!

        while not self.stop_event.is_set():
            try:
                # Wait for the machine to have an assigned PDU port
                if self.machine.pdu_port is None:
                    time.sleep(1)
                    continue
                try:
                    if not session_init():
                        # No jobs for us to run!
                        continue

                    self.log(f"Starting the job: {self.job_config}\n\n", LogLevel.DEBUG)
                    execute_job()
                except Exception:
                    log_exception()
                finally:
                    session_end()
            except Exception:
                # Capture any further exceptions from session_end
                # TODO: Refactor to avoid the cyclomatic complexity.
                # Note: Do not be tempted to log_exception() here! Any
                # exceptions thrown by *that* method could crash our
                # thread.
                traceback.print_exc()

            # TODO: Keep the state of the job in memory for later querying



from datetime import datetime, timedelta
from freezegun import freeze_time
import pytest
from unittest.mock import patch, MagicMock

from server.job import Target, Timeout, Timeouts, ConsoleState, _multiline_string, Deployment, Job

# Target


def test_Target_from_job__no_id_nor_tags():
    with pytest.raises(ValueError) as exc:
        Target.from_job({})

    msg = "The target is neither identified by tags or id. Use empty tags to mean 'any machines'."
    assert msg in str(exc.value)


def test_Target_from_job__id_only():
    target_job = {
        "id": "MyID",
    }

    target = Target.from_job(target_job)
    assert target.id == target_job['id']
    assert target.tags == []
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


def test_Target_from_job__tags_only():
    target_job = {
        "tags": ['tag 1', 'tag 2']
    }

    target = Target.from_job(target_job)
    assert target.id is None
    assert target.tags == target_job['tags']
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


def test_Target_from_job__both_id_and_tags():
    target_job = {
        "id": "MyID",
        "tags": ['tag 1', 'tag 2']
    }

    target = Target.from_job(target_job)
    assert target.id == target_job['id']
    assert target.tags == target_job['tags']
    assert str(target) == f"<Target: id={target.id}, tags={target.tags}>"


# Timeout


def test_Timeout__expiration_test():
    start_time = datetime(2021, 1, 1, 12, 0, 0)
    with freeze_time(start_time.isoformat()):
        timeout = Timeout(name="name", timeout=timedelta(minutes=1), retries=0)
        assert timeout.started_at is None
        assert timeout.active_for is None

        # Start the timeout and check the state
        timeout.start()
        assert timeout.started_at == start_time
        assert timeout.active_for == timedelta()
        assert not timeout.has_expired

        # Go right to the limit of the timeout
        delta = timedelta(seconds=60)
        with freeze_time((start_time + delta).isoformat()):
            assert timeout.started_at == start_time
            assert timeout.active_for == delta
            assert not timeout.has_expired

        # And check that an extra millisecond trip it
        delta = timedelta(seconds=60, milliseconds=1)
        with freeze_time((start_time + delta).isoformat()):
            assert timeout.started_at == start_time
            assert timeout.active_for == delta
            assert timeout.has_expired

        # Stop the timeout and check the state
        timeout.stop()
        assert timeout.started_at is None
        assert timeout.active_for is None


def test_Timeout__retry_lifecycle():
    timeout = Timeout(name="timeout name", timeout=timedelta(seconds=42), retries=1)

    # Check the default state
    assert timeout.started_at is None
    assert timeout.active_for is None
    assert timeout.retried == 0

    # Start the timeout
    start_time = datetime(2021, 1, 1, 12, 0, 0)
    with freeze_time(start_time.isoformat()):
        timeout.start()
        assert timeout.started_at == start_time
        assert timeout.retried == 0
        assert timeout.active_for == timedelta()
        assert not timeout.has_expired

    # Check that the default reset sets started_at to now()
    start_time = datetime(2021, 1, 1, 12, 0, 1)
    with freeze_time(start_time.isoformat()):
        timeout.reset()
        assert timeout.started_at == start_time

        # Check that a resetting to a certain time does as it should
        new_start = start_time - timedelta(seconds=1)
        timeout.reset(new_start)
        assert timeout.started_at == new_start

    # Do the first retry
    assert timeout.retry()
    assert timeout.started_at is None
    assert timeout.retried == 1

    # Second retry should fail
    assert not timeout.retry()


def test_Timeout_from_job():
    delay = timedelta(days=5, hours=6, minutes=7, seconds=8, milliseconds=9)
    timeout = Timeout.from_job("Yeeepeee", {"days": 5, "hours": 6, "minutes": 7,
                                            "seconds": 8, "milliseconds": 9, "retries": 42})

    assert timeout.timeout == delay
    assert timeout.retries == 42
    assert str(timeout) == f"<Timeout Yeeepeee: value={delay}, retries=0/42>"


# Timeouts


def test_Timeouts__overall_with_retries():
    for t_type in [Timeouts.Type.OVERALL.value, Timeouts.Type.INFRA_TEARDOWN.value]:
        with pytest.raises(ValueError) as exc:
            Timeouts({t_type: Timeout("name", timedelta(), retries=1)})
        assert "Neither the overall nor the teardown timeout can have retries" in str(exc.value)


def test_Timeouts__default():
    timeouts = Timeouts({})

    for timeout in timeouts:
        assert timeout.timeout == timedelta.max
        assert timeout.retries == 0

    assert timeouts.expired_list == []
    assert not timeouts.has_expired


def test_Timeouts__expired():
    overall = Timeout("name", timedelta(days=1), retries=0)
    boot_cycle = Timeout("name", timedelta(seconds=0), retries=0)

    overall.start()
    boot_cycle.start()

    timeouts = Timeouts({Timeouts.Type.OVERALL.value: overall, Timeouts.Type.BOOT_CYCLE.value: boot_cycle})
    assert timeouts.has_expired
    assert timeouts.expired_list == [boot_cycle]


def test_Timeouts__from_job():
    job_timeouts = {
        "first_console_activity": {
            "seconds": 45
        },
        "console_activity": {
            "seconds": 13
        }
    }

    timeouts = Timeouts.from_job(job_timeouts)

    assert timeouts.first_console_activity.timeout == timedelta(seconds=45)
    assert timeouts.console_activity.timeout == timedelta(seconds=13)


# ConsoleState


def test_ConsoleState__missing_session_end():
    with pytest.raises(AttributeError):
        ConsoleState(session_end=None, session_reboot=None, job_success=None, job_warn=None)


def test_ConsoleState__simple_lifecycle():
    state = ConsoleState(session_end="session_end", session_reboot=None, job_success=None, job_warn=None)

    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"oh oh oh")
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla session_end blaba\n")
    assert state.job_status == "COMPLETE"
    assert state.session_has_ended
    assert not state.needs_reboot


def test_ConsoleState__lifecycle_with_extended_support():
    state = ConsoleState(session_end="session_end", session_reboot="session_reboot",
                         job_success="job_success", job_warn="job_warn")

    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"oh oh oh")
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla session_reboot blabla")
    assert state.job_status == "INCOMPLETE"
    assert not state.session_has_ended
    assert state.needs_reboot

    state.reset_per_boot_state()
    assert not state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla session_end blaba\n")
    assert state.job_status == "FAIL"
    assert state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla job_success blaba\n")
    assert state.job_status == "PASS"
    assert state.session_has_ended
    assert not state.needs_reboot

    state.process_line(b"blabla job_warn blaba\n")
    assert state.job_status == "WARN"
    assert state.session_has_ended
    assert not state.needs_reboot


def test_ConsoleState_from_job__default():
    console_state = ConsoleState.from_job({})

    assert console_state.session_end == "^\\[[\\d \\.]{12}\\] reboot: Power Down$"
    assert console_state.session_reboot is None
    assert console_state.job_success is None
    assert console_state.job_warn is None


def test_ConsoleState_from_job__full():
    console_state = ConsoleState.from_job({
        "session_end": {
            "regex": "session_end"
        }, "session_reboot": {
            "regex": "session_reboot"
        }, "job_success": {
            "regex": "job_success"
        }, "job_warn": {
            "regex": "job_warn"
        }
    })

    assert console_state.session_end == "session_end"
    assert console_state.session_reboot == "session_reboot"
    assert console_state.job_success == "job_success"
    assert console_state.job_warn == "job_warn"


# _multiline_string


def test_multiline_string():
    assert _multiline_string("toto") == "toto"
    assert _multiline_string(["tag1", "tag2"]) == "tag1 tag2"

    with pytest.raises(AssertionError):
        _multiline_string(1234)


# Deployment


def test_Deployment():
    deployment = Deployment()

    assert deployment.kernel_url is None
    assert deployment.initramfs_url is None
    assert deployment.kernel_cmdline is None

    deployment.update({})

    assert deployment.kernel_url is None
    assert deployment.initramfs_url is None
    assert deployment.kernel_cmdline is None

    deployment.update({"kernel": {"url": "kernel_url", "cmdline": "cmdline"}, "initramfs": {"url": "initrd_url"}})

    assert deployment.kernel_url == "kernel_url"
    assert deployment.initramfs_url == "initrd_url"
    assert deployment.kernel_cmdline == "cmdline"

    assert str(deployment) == """<Deployment:
    kernel_url: kernel_url
    initramfs_url: initrd_url
    kernel_cmdline: cmdline>
"""

# Job


def test_Job__simple():
    simple_job = """
version: 1
target:
  id: "b4:2e:99:f0:76:c5"
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "kernel_url"
      cmdline:
        - my
        - start cmdline
    initramfs:
      url: "initramfs_url"
"""
    job = Job.from_job(simple_job)

    assert job.version == 1
    assert job.deadline == datetime.max

    assert job.target.id == "b4:2e:99:f0:76:c5"
    assert job.target.tags == []

    assert job.deployment_start.kernel_url == "kernel_url"
    assert job.deployment_start.initramfs_url == "initramfs_url"
    assert job.deployment_start.kernel_cmdline == "my start cmdline"

    assert job.deployment_continue.kernel_url == job.deployment_start.kernel_url
    assert job.deployment_continue.initramfs_url == job.deployment_start.initramfs_url
    assert job.deployment_continue.kernel_cmdline == job.deployment_start.kernel_cmdline

    # Make sure the job's __str__ method does not crash
    str(job)


def test_Job__override_continue():
    override_job = """
version: 1
deadline: "2021-03-31 00:00:00"
target:
  id: "b4:2e:99:f0:76:c6"
  tags: ["amdgpu:gfxversion::gfx10"]
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "kernel_url"
      cmdline:
        - my
        - start cmdline
    initramfs:
      url: "initramfs_url"
  continue:
    kernel:
      url: "kernel_url 2"
      cmdline: "my continue cmdline"
    initramfs:
      url: "initramfs_url 2"
"""
    job = Job.from_job(override_job)

    assert job.version == 1
    assert job.deadline == datetime.fromisoformat("2021-03-31 00:00:00")

    assert job.target.id == "b4:2e:99:f0:76:c6"
    assert job.target.tags == ["amdgpu:gfxversion::gfx10"]

    assert job.deployment_start.kernel_url == "kernel_url"
    assert job.deployment_start.initramfs_url == "initramfs_url"
    assert job.deployment_start.kernel_cmdline == "my start cmdline"

    assert job.deployment_continue.kernel_url == "kernel_url 2"
    assert job.deployment_continue.initramfs_url == "initramfs_url 2"
    assert job.deployment_continue.kernel_cmdline == "my continue cmdline"


class MockMachine:
    @property
    def ready_for_service(self):
        return True

    @property
    def id(self):
        return "b4:2e:99:f0:76:c5"

    @property
    def tags(self):
        return ["some", "tags"]

    @property
    def local_tty_device(self):
        return "ttyS0"


class MockBucket:
    @property
    def name(self):
        return "bucket_name"

    @property
    def minio(self):
        return MagicMock(url="minio_url")

    @property
    def credentials(self):
        return MagicMock(dut=("access", "secret"))


@patch('server.config.job_environment_vars')
def test_Job__sample(job_env):
    job_env.return_value = {'MINIO_URL': 'testing-url',
                            'NTP_PEER': '10.42.0.1',
                            'PULL_THRU_REGISTRY': '10.42.0.1:8001'}

    m = MockMachine()
    job = Job.from_path("src/valve_gfx_ci/executor/server/tests/sample_job.yml", m)

    assert job.version == 1
    assert job.deadline == datetime.fromisoformat("2021-03-31 00:00:00")

    assert job.target.id == m.id
    assert job.target.tags == m.tags

    assert job.deployment_start.kernel_url == "testing-url/test-kernel"
    assert job.deployment_start.initramfs_url == "testing-url/test-initramfs"

    assert job.deployment_start.kernel_cmdline == 'b2c.container="docker://10.42.0.1:8001/infra/machine_registration:latest check" b2c.ntp_peer="10.42.0.1" b2c.pipefail b2c.cache_device=auto b2c.container="-v /container/tmp:/storage docker://10.42.0.1:8002/tests/mesa:12345" console=ttyS0,115200 earlyprintk=vga,keep SALAD.machine_id=b4:2e:99:f0:76:c5'  # noqa: E501

    assert job.deployment_continue.kernel_url == "testing-url/test-kernel"
    assert job.deployment_continue.initramfs_url == "testing-url/test-initramfs"
    assert job.deployment_continue.kernel_cmdline == 'b2c.container="docker://10.42.0.1:8001/infra/machine_registration:latest check" b2c.ntp_peer=10.42.0.1 b2c.pipefail b2c.cache_device=auto b2c.container="-v /container/tmp:/storage docker://10.42.0.1:8002/tests/mesa:12345 resume"'  # noqa: E501


def test_Job__invalid_format():
    job = """
version: 1
target:
  id: "b4:2e:99:f0:76:c6"
console_patterns:
  session_end:
    regex: "session_end"
  reboot:
    regex: "toto"
deployment:
  start:
    kernel:
      url: "kernel_url"
      cmdline:
        - my
        - start cmdline
    initramfs:
      url: "initramfs_url"
"""

    with pytest.raises(ValueError) as exc:
        Job.from_job(job)

    assert str(exc.value) == "{'console_patterns': {'reboot': ['Unknown field.']}}"


@patch('server.config.job_environment_vars')
def test_Job__from_machine(job_env):
    job_env.return_value = {'NTP_PEER': '10.42.0.1'}

    simple_job = """
version: 1
target:
  id: {{ machine_id }}
console_patterns:
  session_end:
    regex: "session_end"
deployment:
  start:
    kernel:
      url: "kernel_url"
      cmdline:
        - my {{ minio_url }}
        - start cmdline {{ ntp_peer }}
    initramfs:
      url: "initramfs_url"
"""
    job = Job.render_with_resources(simple_job, MockMachine(), MockBucket())

    assert job.version == 1
    assert job.deadline == datetime.max

    assert job.target.id == "b4:2e:99:f0:76:c5"
    assert job.target.tags == []

    assert job.deployment_start.kernel_url == "kernel_url"
    assert job.deployment_start.initramfs_url == "initramfs_url"
    assert job.deployment_start.kernel_cmdline == "my minio_url start cmdline 10.42.0.1"

    assert job.deployment_continue.kernel_url == job.deployment_start.kernel_url
    assert job.deployment_continue.initramfs_url == job.deployment_start.initramfs_url
    assert job.deployment_continue.kernel_cmdline == job.deployment_start.kernel_cmdline



from collections import namedtuple
import glob
import jinja2
import os
import re
import requests
import shutil
import signal
import subprocess
import sys
import tempfile
import time

from . import config
from .logger import logger

DEFAULT_CONFIG_PATHS = {
    'BOOTS_ROOT': '/boots',
    'TFTP_DIR': '/boots/tftp',
    'PXELINUX_CONFIG_DIR': '/boots/tftp/pxelinux.cfg',
}

SYSLINUX_URL = 'https://mirrors.edge.kernel.org/pub/linux/utils/boot/syslinux/syslinux-6.03.tar.gz'
BASE_DIR = os.path.dirname(__file__)

JINJA_ENVIRONMENT = jinja2.Environment(
    loader=jinja2.FileSystemLoader(os.path.join(BASE_DIR, 'templates')),
    extensions=['jinja2.ext.autoescape'],
    autoescape=False,
    trim_blocks=True,
    lstrip_blocks=True)
Host = namedtuple('Host', ['mac_addr', 'ip_addr', 'hostname'])


def render_template(tmpl_name, **options):
    return JINJA_ENVIRONMENT.get_template(tmpl_name).render(**options)


def split_mac_addr(s):
    """Return an array of bytes for the mac address in string s. This
    method is flexible in its input."""
    s = s.lower()
    m = re.match("[0-9a-f]{2}([-:]?)[0-9a-f]{2}(\\1[0-9a-f]{2}){4}$",
                 s)
    if not m:
        raise ValueError(f"{s} is not a valid mac address")
    delim = m.groups()[0]
    if delim:
        return s.split(delim)
    else:
        return [s[i:i+2] for i in range(0, len(s), 2)]


def create_if_not_exists(filename):
    if not os.path.exists(filename):
        with open(filename, 'w') as _:
            pass


def parse_dhcp_hosts(s):
    hosts = []
    for host_desc in s.splitlines():
        host_desc = host_desc.strip()
        if not len(host_desc) or host_desc.startswith("#"):
            continue
        m = re.match(
            r'^(?P<mac>[^,]+),(?P<ip>[^,]+),set:(?P<hostname>.+)$',
            host_desc)
        if not m:
            raise RuntimeError("Malformed DHCP hosts file")
        hosts.append(Host(*m.groups()))
    return hosts


def provision_network_boot_service(default_kernel_path,
                                   default_initrd_path,
                                   default_kernel_cmdline,
                                   config_paths):  # pragma: nocover
    os.makedirs(config_paths['TFTP_DIR'], exist_ok=True)
    if not (os.path.exists(os.path.join(config_paths['TFTP_DIR'], 'lpxelinux.0')) or
            os.path.exists(os.path.join(config_paths['TFTP_DIR'], 'ldlinux.c32'))):
        r = requests.get(SYSLINUX_URL, stream=True)
        with (tempfile.NamedTemporaryFile() as tf,
              tempfile.TemporaryDirectory() as td):
            logger.info("Downloading PXELINUX files...")
            for chunk in r.iter_content(chunk_size=16*1024):
                tf.write(chunk)
            subprocess.check_call(["tar", "xzf",
                                   tf.name, "-C", td])
            for f in [
                    'syslinux-6.03/bios/core/lpxelinux.0',
                    'syslinux-6.03/bios/com32/elflink/ldlinux/ldlinux.c32',
                    'syslinux-6.03/efi64/efi/syslinux.efi',
                    'syslinux-6.03/efi64/com32/elflink/ldlinux/ldlinux.e64']:
                shutil.move(os.path.join(td, f),
                            os.path.join(config_paths['TFTP_DIR'],
                                         os.path.basename(f)))
    os.makedirs(config_paths['PXELINUX_CONFIG_DIR'], exist_ok=True)
    default_pxelinux_config = os.path.join(config_paths['PXELINUX_CONFIG_DIR'], 'default')
    if not os.path.isfile(default_pxelinux_config):
        with open(default_pxelinux_config, 'w') as f:
            f.write(f"""DEFAULT def
label def
KERNEL {default_kernel_path}
INITRD {default_initrd_path}
APPEND {default_kernel_cmdline}
""")


class Dnsmasq():
    def __init__(self,
                 private_interface: str,
                 config_paths):
        if not shutil.which('dnsmasq'):
            raise RuntimeError("No dnsmasq found on the system!")

        self.config_paths = config_paths
        self.pid_file = os.path.join(config_paths['BOOTS_ROOT'], 'dnsmasq.pid')
        self.leases_file = os.path.join(config_paths['BOOTS_ROOT'], 'dnsmasq.leases')

        self.options_file = os.path.join(config_paths['BOOTS_ROOT'], 'options.dhcp')
        self.hosts_file = os.path.join(config_paths['BOOTS_ROOT'], 'hosts.dhcp')

        if not os.path.isfile(self.options_file):
            with open(self.options_file, 'w') as f:
                f.write("""
# Not tested, but interesting hook point for future DHCP options
option:ntp-server,10.42.0.1
""")
        create_if_not_exists(self.hosts_file)

        if not config.BOOTS_DISABLE_DNSMASQ:
            self.dnsmasq = subprocess.Popen(
                [
                    'dnsmasq',
                    '--port=0',
                    f'--pid-file={self.pid_file}',
                    f'--dhcp-hostsfile={self.hosts_file}',
                    f'--dhcp-optsfile={self.options_file}',
                    f'--dhcp-leasefile={self.leases_file}',
                    '--dhcp-match=set:efi-x86_64,option:client-arch,7',
                    '--dhcp-boot=tag:efi-x86_64,syslinux.efi',
                    '--dhcp-boot=lpxelinux.0',
                    '--dhcp-range=10.42.0.10,10.42.0.100',
                    '--dhcp-script=/bin/echo',
                    # f'--dhcp-hostsfile={static_hosts_file}',
                    f'--enable-tftp={private_interface}',
                    f'--tftp-root={config_paths["BOOTS_ROOT"]}/tftp',
                    # TODO: Rotation
                    f'--log-facility={config_paths["BOOTS_ROOT"]}/dnsmasq.log',
                    '--log-queries=extra',
                    '--conf-file=/dev/null',
                    f'--interface={private_interface}'
                ],
                bufsize=0,
            )

            # We don't want to return in a not-ready state.
            self._wait_for_dnsmasq_to_fork()

    def _wait_for_dnsmasq_to_fork(self):  # pragma: nocover
        did_fork = False
        for _ in range(10):
            logger.debug("Waiting for dnsmasq to fork...")
            if os.path.exists(self.pid_file):
                with open(self.pid_file) as f:
                    pid = f.read().strip()
                    if os.path.isfile(f'/proc/{pid}/status'):
                        did_fork = True
                        break
            time.sleep(0.2)
        if did_fork:
            logger.debug("dnsmasq is ready")
        else:
            logger.error("dnsmasq did not start in time")
            sys.exit(1)

    def _pid(self):  # pragma: nocover
        with open(self.pid_file) as pidfile:
            return int(pidfile.read())

    def reload(self):  # pragma: nocover
        if not config.BOOTS_DISABLE_DNSMASQ:
            os.kill(self._pid(), signal.SIGHUP)

    def stop(self):  # pragma: nocover
        if not config.BOOTS_DISABLE_DNSMASQ:
            os.kill(self._pid(), signal.SIGKILL)
            os.remove(self.pid_file)

    def add_static_address(self, mac_addr, ip_addr, hostname):
        mac_addr = ':'.join(split_mac_addr(mac_addr))

        logger.info("%s. ip=%s hostname=%s",
                    mac_addr, ip_addr, hostname)

        with open(self.hosts_file, 'r') as f:
            machines = parse_dhcp_hosts(f.read())
            new_machines = [m for m in machines if m.mac_addr != mac_addr]
            new_machines.append(Host(mac_addr, ip_addr, hostname))
            config = render_template(
                'dnsmasq-dhcp-host.jinja',
                machines=sorted(new_machines, key=lambda m: m[0]))
        try:
            # Make an effort to avoid dnsmasq seeing a partial write.
            with tempfile.NamedTemporaryFile(mode='w',
                                             dir=self.config_paths['PXELINUX_CONFIG_DIR'],
                                             delete=False) as tf:
                tf.write(config)
                tf.flush()
                os.rename(tf.name, self.hosts_file)
                # Ensure the dnsmasq nobody can read us
                os.chmod(self.hosts_file, 0o644)
        finally:
            self.reload()


class BootService():
    def __init__(self,
                 private_interface=None,
                 default_kernel=None,
                 default_initrd=None,
                 default_cmdline=None,
                 config_paths=DEFAULT_CONFIG_PATHS):
        self.private_interface = private_interface or config.PRIVATE_INTERFACE
        self.default_kernel = default_kernel or config.BOOTS_DEFAULT_KERNEL
        self.default_initrd = default_initrd or config.BOOTS_DEFAULT_INITRD
        self.default_cmdline = default_cmdline or config.BOOTS_DEFAULT_CMDLINE
        self.config_paths = config_paths

        # Ensure these steps happen on the calling thread, not
        provision_network_boot_service(
            self.default_kernel,
            self.default_initrd,
            self.default_cmdline,
            config_paths)

        self.dnsmasq = Dnsmasq(self.private_interface, config_paths)

    def stop(self):
        self.dnsmasq.stop()
        for config_file in glob.glob(os.path.join(self.config_paths['PXELINUX_CONFIG_DIR'], '01-*')):
            os.remove(config_file)

    def write_network_config(self, mac_address, ip_address, hostname):
        # Are static mappings that useful here? We're basically
        # enforcing the initial DHCP offer, but it does allow a
        # specific address to be set at least.
        logger.debug("%s: ip=%s hostname=%s",
                     mac_address, ip_address, hostname)
        self.dnsmasq.add_static_address(mac_address, ip_address, hostname)

    def encode_pxelinux_filename(self, mac_addr):
        if isinstance(mac_addr, str):
            mac_addr = split_mac_addr(mac_addr)
        assert isinstance(mac_addr, list) and \
            len(mac_addr) == 6
        return "01-" + '-'.join(mac_addr)

    def pxelinux_filename_from_mac(self, mac_addr):
        if mac_addr == "default":
            filename = "default"
        else:
            filename = self.encode_pxelinux_filename(mac_addr)

        return os.path.join(self.config_paths['PXELINUX_CONFIG_DIR'],
                            filename)

    def write_pxelinux_config(self,
                              mac_addr,
                              kernel_path=None,
                              cmdline=None,
                              initrd_path=None):
        pxelinux_filename = self.pxelinux_filename_from_mac(mac_addr)
        logger.info("%s. kernel=%s, cmdline=%s, initrd=%s",
                    pxelinux_filename, kernel_path, cmdline, initrd_path)
        result = (render_template(
            'pxelinux-raw.jinja',
            kernel_path=kernel_path or self.default_kernel,
            cmdline=cmdline or self.default_cmdline,
            initrd_path=initrd_path or self.default_initrd))

        with open(pxelinux_filename, 'w') as config:
            config.write(result)

    def remove_pxelinux_config(self, mac_addr):
        pxelinux_filename = self.pxelinux_filename_from_mac(mac_addr)
        if os.path.exists(pxelinux_filename):
            logger.info("%s", pxelinux_filename)
            os.remove(pxelinux_filename)


if __name__ == '__main__':  # pragma: nocover
    if len(sys.argv) == 2:
        private_interface = sys.argv[1]
    else:
        private_interface = config.PRIVATE_INTERFACE
    try:
        boots = BootService(private_interface=private_interface, config_paths=DEFAULT_CONFIG_PATHS)
        boots.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        boots.stop()



from unittest.mock import MagicMock, patch, PropertyMock
import pytest
import copy
import time

from server.pdu import PDUState, PDUPort, PDU
from server.pdu.drivers.apc import ApcMasterswitchPDU
from server.pdu.drivers.cyberpower import PDU41004
from server.pdu.drivers.dummy import DummyPDU
from server.pdu.drivers.snmp import (
    retry_on_known_errors,
    SnmpPDU,
    ManualSnmpPDU,
    snmp_get,
    snmp_set,
    snmp_walk
)


def test_PDUState_UNKNOW_is_invalid_action():
    assert PDUState.UNKNOWN not in PDUState.valid_actions()
    assert PDUState.UNKNOWN.is_valid_action is False


def test_PDUState_valid_actions_contain_basics():
    for action in ["ON", "OFF", "REBOOT"]:
        assert action in [s.name for s in PDUState.valid_actions()]
        assert getattr(PDUState, action).is_valid_action is True


def test_PDUPort_get_set():
    pdu = MagicMock(get_port_state=MagicMock(return_value=PDUState.OFF))
    port = PDUPort(pdu, 42, label="My Port")
    assert port.label == "My Port"

    last_shutdown = port.last_shutdown

    pdu.set_port_state.assert_not_called()
    pdu.get_port_state.assert_not_called()

    assert pdu.set_port_state.call_count == 0
    port.set(PDUState.ON)
    pdu.set_port_state.assert_called_with(42, PDUState.ON)
    assert pdu.set_port_state.call_count == 1
    assert port.last_shutdown == last_shutdown

    assert port.state == PDUState.OFF
    pdu.get_port_state.assert_called_with(42)

    # Check that setting the port to the same value does not change anything
    port.set(PDUState.OFF)
    assert pdu.set_port_state.call_count == 1
    assert port.last_shutdown == last_shutdown

    # Check that whenever we set the PDU state to OFF, we update last_shutdown
    pdu.get_port_state.return_value = PDUState.ON
    port.set(PDUState.OFF)
    assert port.last_shutdown > last_shutdown


def test_PDUPort_eq():
    params = {
        "pdu": "pdu",
        "port_id": 42,
        "label": "label",
        "min_off_time": "min_off_time"
    }

    assert PDUPort(**params) == PDUPort(**params)
    for param in params:
        n_params = dict(params)
        n_params[param] = "modified"
        assert PDUPort(**params) != PDUPort(**n_params)


def test_PDU_defaults():
    pdu = PDU("MyPDU")

    assert pdu.name == "MyPDU"
    assert pdu.ports == []
    assert pdu.set_port_state(42, PDUState.ON) is False
    assert pdu.get_port_state(42) == PDUState.UNKNOWN


def test_PDU_supported_pdus():
    pdus = PDU.supported_pdus()
    assert "dummy" in pdus


def test_PDU_create():
    pdu = PDU.create("dummy", "name", {})
    assert pdu.name == "name"

    with pytest.raises(ValueError):
        PDU.create("invalid", "name", {})


# Drivers

@pytest.fixture(autouse=True)
def reset_easysnmp_mock(monkeypatch):
    import server.pdu.drivers.snmp as snmp

    global snmp_get, snmp_set, snmp_walk, time_sleep
    m1, m2, m3, m4 = MagicMock(), MagicMock(), MagicMock(), MagicMock()
    # REVIEW: I wonder if there's a clever way of covering the
    # difference in import locations between here and snmp.py
    monkeypatch.setattr(snmp, "snmp_walk", m1)
    monkeypatch.setattr(snmp, "snmp_get", m2)
    monkeypatch.setattr(snmp, "snmp_set", m3)
    monkeypatch.setattr(time, "sleep", m4)
    snmp_walk = m1
    snmp_get = m2
    snmp_set = m3
    time_sleep = m4


@patch("random.random", return_value=0.42)
def test_driver_BaseSnmpPDU_retry_on_known_errors__known_error(random_mock):
    global retriable_error_call_count
    retriable_error_call_count = 0

    @retry_on_known_errors
    def retriable_error():
        global retriable_error_call_count

        assert time_sleep.call_count == retriable_error_call_count

        retriable_error_call_count += 1
        raise SystemError("<built-in function set> returned NULL without setting an error")

    with pytest.raises(ValueError):
        retriable_error()

    time_sleep.assert_called_with(1.42)
    assert time_sleep.call_count == retriable_error_call_count
    assert retriable_error_call_count == 3


def test_driver_BaseSnmpPDU_retry_on_known_errors__unknown_error():
    global unretriable_error_call_count
    unretriable_error_call_count = 0

    @retry_on_known_errors
    def unretriable_error():
        global unretriable_error_call_count
        unretriable_error_call_count += 1
        raise SystemError("Unknown error")

    with pytest.raises(SystemError):
        unretriable_error()

    time_sleep.assert_not_called()
    assert unretriable_error_call_count == 1


class MockSnmpPDU(SnmpPDU):
    system_id = '1234.1.2.3.4'
    outlet_labels = '4.5.4.5'
    outlet_status = '4.5.4.6'

    state_mapping = {
        PDUState.ON: 2,
        PDUState.OFF: 3,
        PDUState.REBOOT: 4,
    }

    state_transition_delay_seconds = 5


def test_driver_SnmpPDU_eq():
    params = {
        "hostname": "hostname",
        "community": "community"
    }

    assert MockSnmpPDU("name", params) == MockSnmpPDU("name", params)
    for param in params:
        n_params = dict(params)
        n_params[param] = "modified"
        assert MockSnmpPDU("name", params) != MockSnmpPDU("name", n_params)


def test_driver_SnmpPDU_listing_ports():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    snmp_walk.return_value = [MagicMock(value="P1"), MagicMock(value="P2")]
    snmp_walk.assert_not_called()
    ports = pdu.ports
    snmp_walk.assert_called_with(pdu.outlet_labels_oid,
                                 hostname=pdu.hostname, community=pdu.community,
                                 version=1)

    # Check that the labels are stored, and the port IDs are 1-indexed
    for i in range(0, 2):
        assert ports[i].port_id == i+1
        assert ports[i].label == f"P{i+1}"

    snmp_walk.side_effect = SystemError("<built-in function walk> returned NULL without setting an error")
    with pytest.raises(ValueError):
        pdu.ports


def test_driver_BaseSnmpPDU_port_label_mapping():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    snmp_walk.return_value = [
        MagicMock(value="P1"),
        MagicMock(value="P2")
    ]
    snmp_set.return_value = True
    assert pdu.set_port_state("P1", PDUState.REBOOT) is True
    snmp_set.assert_called_with(pdu.outlet_ctrl_oid(1),
                                pdu.state_mapping[PDUState.REBOOT], 'i',
                                hostname=pdu.hostname, community=pdu.community,
                                version=1)
    assert pdu.set_port_state("P2", PDUState.REBOOT) is True
    snmp_set.assert_called_with(pdu.outlet_ctrl_oid(2),
                                pdu.state_mapping[PDUState.REBOOT], 'i',
                                hostname=pdu.hostname, community=pdu.community,
                                version=1)
    with pytest.raises(ValueError):
        pdu.set_port_state("flubberbubber", PDUState.OFF)


def test_driver_BaseSnmpPDU_get_port():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    type(snmp_get.return_value).value = \
        PropertyMock(return_value=pdu.state_mapping[PDUState.REBOOT])
    snmp_get.assert_not_called()
    pdu_state = pdu.get_port_state(2)
    assert pdu_state == PDUState.REBOOT
    snmp_get.assert_called_with(pdu.outlet_status_oid(2),
                                hostname=pdu.hostname, community=pdu.community,
                                version=1)

    snmp_get.side_effect = SystemError("<built-in function get> returned NULL without setting an error")
    with pytest.raises(ValueError):
        pdu.get_port_state(2)


def test_driver_BaseSnmpPDU_set_port():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})
    type(snmp_get.return_value).value = \
        PropertyMock(return_value=pdu.state_mapping[PDUState.REBOOT])
    snmp_set.return_value = True
    snmp_set.assert_not_called()
    assert pdu.set_port_state(2, PDUState.REBOOT) is True
    snmp_set.assert_called_with(pdu.outlet_ctrl_oid(2),
                                pdu.state_mapping[PDUState.REBOOT], 'i',
                                hostname=pdu.hostname, community=pdu.community,
                                version=1)

    snmp_set.side_effect = SystemError("<built-in function set> returned NULL without setting an error")
    with pytest.raises(ValueError):
        pdu.set_port_state(2, PDUState.REBOOT)


def test_driver_BaseSnmpPDU_action_translation():
    pdu = MockSnmpPDU("MyPDU", {"hostname": "127.0.0.1"})

    # Check the state -> SNMP value translation
    for action in PDUState.valid_actions():
        assert pdu.inverse_state_mapping[pdu.state_mapping[action]] == action

    with pytest.raises(KeyError):
        pdu.state_mapping[PDUState.UNKNOWN]


def test_driver_ApcMasterswitchPDU_check_OIDs():
    pdu = ApcMasterswitchPDU("MyPDU", {"hostname": "127.0.0.1"})

    assert pdu.outlet_labels_oid == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.4"
    assert pdu.outlet_status_oid(10) == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.3.10"
    assert pdu.outlet_ctrl_oid(10) == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.3.10"


def test_driver_PDU41004_check_OIDs():
    pdu = PDU41004("MyPDU", {"hostname": "127.0.0.1"})
    assert pdu.outlet_labels_oid == f"{pdu.oid_enterprise}.3808.1.1.3.3.3.1.1.2"
    assert pdu.outlet_status_oid(10) == f"{pdu.oid_enterprise}.3808.1.1.3.3.3.1.1.4.10"
    assert pdu.outlet_ctrl_oid(10) == f"{pdu.oid_enterprise}.3808.1.1.3.3.3.1.1.4.10"


def test_driver_DummyPDU():
    ports = ['P1', 'P2', 'P3']
    pdu = DummyPDU("MyPDU", {"ports": ports})

    assert [p.label for p in pdu.ports] == ports
    assert pdu.get_port_state(0) == PDUState.ON
    pdu.set_port_state(0, PDUState.OFF)
    assert pdu.get_port_state(0) == PDUState.OFF


def test_driver_ManualSnmpPDU_check_OIDs_and_default_actions():
    pdu = ManualSnmpPDU("MyPDU", config={
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    })

    assert pdu.community == "private"
    assert pdu.outlet_labels == "5.6.7.8"
    assert pdu.outlet_status_oid(10) == "1.3.6.1.4.1.1.2.3.4.5.6.7.9.10"
    assert pdu.outlet_ctrl_oid(10) == "1.3.6.1.4.1.1.2.3.4.5.6.7.10.10"
    assert pdu.state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])
    assert pdu.inverse_state_mapping.keys() == set([1, 2, 3])
    for k, _ in pdu.state_mapping.items():
        assert pdu.inverse_state_mapping[pdu.state_mapping[k]] == k


def test_driver_ManualSnmpPDU_invalid_actions():
    with pytest.raises(ValueError):
        ManualSnmpPDU("MyPDU", config={
            "hostname": "127.0.0.1",
            "system_id": "1.2.3.4",
            "outlet_labels": "5.6.7.8",
            "outlet_status": "5.6.7.9",
            "outlet_ctrl": "5.6.7.10",
            "state_mapping": {
                "on": "FUDGE",
                "off": 2,
                "reboot": 3,
            },
        })


def test_driver_ManualSnmpPDU_missing_actions():
    with pytest.raises(AssertionError):
        ManualSnmpPDU("MyPDU", config={
            "hostname": "127.0.0.1",
            "system_id": "1.2.3.4",
            "outlet_labels": "5.6.7.8",
            "outlet_status": "5.6.7.9",
            "outlet_ctrl": "5.6.7.10",
            "state_mapping": {
                "off": 2,
                "reboot": 3,
            },
        })


def test_driver_ManualSnmpPDU_missing_parameters():
    valid_config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    }

    ManualSnmpPDU("MyPDU", config=valid_config)
    for required_param in ["hostname", "system_id", "outlet_labels", "outlet_status", "state_mapping"]:
        new_config = copy.deepcopy(valid_config)
        del new_config[required_param]
        with pytest.raises((ValueError, AssertionError)):
            ManualSnmpPDU("MyPDU", config=new_config)


def test_driver_ManualSnmpPDU_weird_inverses():
    valid_config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
        "inverse_state_mapping": {
            "on": 2,
            "off": 3,
            "reboot": 4,
        },
    }

    pdu = ManualSnmpPDU("MyPDU", config=valid_config)
    for k, _ in pdu.state_mapping.items():
        assert pdu.inverse_state_mapping[pdu.state_mapping[k]] == k + 1



from .. import logger, PDU, PDUPort, PDUState

import time
import socket
from contextlib import contextmanager


class VirtualPDU(PDU):  # pragma: nocover
    def __init__(self, name, config):
        logger.info("Creating a virtual PDU named %s", name)
        self.host, self.port = config.get('hostname', 'localhost:9191').split(':')
        self.port = int(self.port)
        logger.info("Connecting to %s:%d", self.host, self.port)
        with self.conn() as s:
            s.sendall((0).to_bytes(4, byteorder='big'))
            num_ports = int(s.recv(1)[0])
            self._ports = [PDUPort(self, i) for i in range(num_ports)]
        super().__init__(name)

    @property
    def ports(self):
        return self._ports

    @contextmanager
    def conn(self, *args, **kw):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        s.connect((self.host, self.port))
        try:
            yield s
        finally:
            s.close()

    def set_port_state(self, port_id, state):
        if state == PDUState.OFF or state == PDUState.REBOOT:
            cmd = (port_id << 2) | 0x02
            with self.conn() as s:
                s.sendall(cmd.to_bytes(4, byteorder='big'))
                assert(s.recv(1)[0] == 0x01)
        if self.state_transition_delay_seconds:
            time.sleep(self.state_transition_delay_seconds)
        if state == PDUState.ON or state == PDUState.REBOOT:
            cmd = (port_id << 2) | 0x01
            with self.conn() as s:
                s.sendall(cmd.to_bytes(4, byteorder='big'))
                assert(s.recv(1)[0] == 0x01)
        if self.state_transition_delay_seconds:
            time.sleep(self.state_transition_delay_seconds)

    def get_port_state(self, port_id):
        cmd = port_id << 2 | 0x03
        with self.conn() as s:
            s.sendall(cmd.to_bytes(4, byteorder='big'))
            state = int(s.recv(1)[0])
            logger.info("port state %x", state)
            if state == 0x03:
                return PDUState.ON
            elif state == 0x04:
                return PDUState.OFF
            elif state == 0x05:
                return PDUState.UNKNOWN
            else:
                assert(False)



from dataclasses import dataclass
from jinja2 import Template
import requests

from .logger import logger
from . import config


@dataclass
class GitlabRunnerRegistration:
    id: int
    token: str


def register_runner(gitlab_url: str, registration_token: str,
                    description: str, tag_list: list[str],
                    run_untagged: bool = False, maximum_timeout: int = 3600):
    params = {
        "token": registration_token,
        "description": description,
        "tag_list": ",".join(tag_list),
        "run_untagged": run_untagged,
        "maximum_timeout": maximum_timeout
    }

    r = requests.post(f"{gitlab_url}/api/v4/runners", params=params)
    if r.status_code == 201:
        return GitlabRunnerRegistration(**r.json())
    else:
        return None


def unregister_runner(gitlab_url: str, token: str):
    requests.delete(f"{gitlab_url}/api/v4/runners", params={"token": token})


def verify_runner_token(gitlab_url: str, token: str):
    r = requests.post(f"{gitlab_url}/api/v4/runners/verify", params={"token": token})
    return r.status_code == 200


def generate_runner_config(mars_db):
    logger.info("Generate the GitLab runner configuration")
    with open(config.GITLAB_CONF_TEMPLATE_FILE) as f:
        config_toml = Template(f.read()).render(config=config, mars_db=mars_db)

    with open(config.GITLAB_CONF_FILE, 'w') as f:
        f.write(config_toml)



import os
from typing import Dict

configurables = {
    'EXECUTOR_HOST': '0.0.0.0',
    'EXECUTOR_PORT': 80,
    'EXECUTOR_REGISTRATION_JOB': '/app/job_templates/register.yml.j2',
    'EXECUTOR_BOOTLOOP_JOB': '/app/job_templates/bootloop.yml.j2',
    'SERGENT_HARTMAN_BOOT_COUNT': '100',
    'SERGENT_HARTMAN_QUALIFYING_BOOT_COUNT': '100',
    'SERGENT_HARTMAN_REGISTRATION_RETRIAL_DELAY': '120',
    'GITLAB_URL': 'https://gitlab.freedesktop.org',
    'GITLAB_CONF_FILE': '/etc/gitlab-runner/config.toml',
    'GITLAB_CONF_TEMPLATE_FILE': '/app/templates/gitlab_runner_config.toml.j2',
    'FARM_NAME': None,
    'MARS_DB_FILE': '/app/config/mars.yaml',
    'SALAD_URL': 'http://10.42.0.1:8005',
    'BOOTS_ROOT': '/boots',
    'BOOTS_TFTP_ROOT': '/boots/tftp',
    'BOOTS_PXELINUX_CONFIG_DIR': '/boots/tftp/pxelinux.cfg',
    'BOOTS_DISABLE_DNSMASQ': None,
    'MINIO_URL': 'http://10.42.0.1:9000',
    'MINIO_ROOT_USER': 'minioadmin',
    'MINIO_ROOT_PASSWORD': 'minio-root-password',
    'MINIO_ADMIN_ALIAS': 'local',
    'FDO_PROXY_REGISTRY': '10.42.0.1:8002',
    'LOCAL_REGISTRY': '10.42.0.1:8004',
    'PRIVATE_INTERFACE': 'private',
    'BOOTS_DEFAULT_KERNEL': 'http://10.42.0.1:9000/boot/default_kernel',
    'BOOTS_DEFAULT_INITRD': 'http://10.42.0.1:9000/boot/default_boot2container.cpio.xz',
    'BOOTS_DEFAULT_CMDLINE': 'b2c.container="-ti --tls-verify=false docker://10.42.0.1:8002/mupuf/valve-infra/machine_registration:latest register" b2c.ntp_peer="10.42.0.1" b2c.cache_device=none loglevel=6'  # noqa
}

__all__ = []


for config_option, default in configurables.items():
    globals()[config_option] = os.environ.get(config_option,
                                              default)
    __all__.append(config_option)


def job_environment_vars() -> Dict[str, str]:  # pragma: nocover
    """Return environment variables useful for job submission as a
    dictionary."""
    return {
        k: globals()[k] for k in [
            'MINIO_URL',
            'FDO_PROXY_REGISTRY',
            'LOCAL_REGISTRY',
        ]
    }



from .. import PDU, PDUPort, PDUState


class DummyPDU(PDU):
    def __init__(self, name, config):
        self._ports = []

        for i, port_label in enumerate(config.get('ports', [])):
            port = PDUPort(self, i, port_label)
            port._state = PDUState.ON
            self._ports.append(port)

        super().__init__(name)

    @property
    def ports(self):
        return self._ports

    def set_port_state(self, port_id, state):
        port = self.ports[int(port_id)]
        port._state = state

    def get_port_state(self, port_id):
        port = self.ports[int(port_id)]
        return port._state



from datetime import datetime
from enum import IntEnum

import time
from logging import getLogger, getLevelName, Formatter, StreamHandler


logger = getLogger(__name__)
logger.setLevel(getLevelName('DEBUG'))
log_formatter = \
    Formatter("%(asctime)s [%(threadName)s] [%(levelname)s] %(funcName)s: "
              "%(message)s")
console_handler = StreamHandler()
console_handler.setFormatter(log_formatter)
logger.addHandler(console_handler)


class PDUState(IntEnum):
    UNKNOWN = 0
    OFF = 1
    ON = 2
    REBOOT = 3

    @classmethod
    def valid_actions(cls):
        return [s for s in PDUState if s.value > 0]

    @property
    def is_valid_action(self):
        return self in self.valid_actions()


class PDUPort:
    def __init__(self, pdu, port_id, label=None, min_off_time=5):
        self.pdu = pdu
        self.port_id = port_id
        self.label = label
        self.min_off_time = min_off_time

        self.last_shutdown = datetime.now()

    def set(self, state):
        # Check the current state before writing it
        cur_state = self.state
        logger.debug("set: %s -> %s", cur_state, state)
        if cur_state == state:
            return

        if cur_state == PDUState.OFF and state == PDUState.ON:
            # Enforce a minimum amount of time between state changes
            time_spent_off = (datetime.now() - self.last_shutdown).total_seconds()
            if time_spent_off < self.min_off_time:
                time.sleep(self.min_off_time - time_spent_off)

        self.pdu.set_port_state(self.port_id, state)

        if state == PDUState.OFF:
            self.last_shutdown = datetime.now()

    @property
    def state(self):
        return self.pdu.get_port_state(self.port_id)

    def __eq__(self, other):
        for attr in ["pdu", "port_id", "label", "min_off_time"]:
            if getattr(self, attr, None) != getattr(other, attr, None):
                return False
        return True


class PDU:
    def __init__(self, name):
        self.name = name
        # FIXME: It would be better to poll on state changes than wait
        # arbitrary amounts of time. Abstracting state transitions
        # into a method, so that subclasses can do more intelligent
        # things (in SNMP, a better thing can be done, for instance),
        # would be a start. There's a couple things in play for these pauses,
        #  1. The firmware on different PDUs needs differing amounts
        #  of time to settle in state transitions.  2. SNMP PDUs
        #  typically offer a way for the user to add state transition
        #  delays of their own, and these come with defaults that we
        #  need to wait *at least* that long for. We should do a
        #  better job of provisioning these values to suit our needs.
        if not hasattr(self, 'state_transition_delay_seconds'):
            self.state_transition_delay_seconds = 1

    @property
    def ports(self):
        # NOTICE: Left for drivers to implement
        return []

    def set_port_state(self, port_id, state):
        # NOTICE: Left for drivers to implement
        return False

    def get_port_state(self, port_id):
        # NOTICE: Left for drivers to implement
        return PDUState.UNKNOWN

    @classmethod
    def supported_pdus(cls):
        from .drivers.apc import ApcMasterswitchPDU
        from .drivers.cyberpower import PDU41004, PDU15SWHVIEC12ATNET
        from .drivers.dummy import DummyPDU
        from .drivers.virtual import VirtualPDU
        from .drivers.snmp import SnmpPDU

        return {
            "apc_masterswitch": ApcMasterswitchPDU,
            "cyberpower_pdu41004": PDU41004,
            "cyberpower_pdu15swhviec12atnet": PDU15SWHVIEC12ATNET,
            "dummy": DummyPDU,
            "vpdu": VirtualPDU,
            "snmp": SnmpPDU,
        }

    @classmethod
    def create(cls, model_name, pdu_name, config):
        driver = cls.supported_pdus().get(model_name)

        if driver is None:
            raise ValueError(f"Unknown model name '{model_name}'")

        return driver(pdu_name, config)



from unittest.mock import MagicMock
from datetime import datetime
import logging
import base64

from server.message import Message, MessageType, ControlMessage, JobIOMessage, SessionEndMessage, JobStatus
import pytest


def test_JobStatus():
    for status in JobStatus:
        assert status.status_code == status.value
        assert JobStatus.from_str(status.name) == status


def test_MessageType_message_class():
    assert MessageType.CONTROL.message_class == ControlMessage
    assert MessageType.JOB_IO.message_class == JobIOMessage
    assert MessageType.SESSION_END.message_class == SessionEndMessage


def test_ControlMessage():
    message = "Hello world"
    severity = logging.CRITICAL
    msg = ControlMessage.create(message=message, severity=severity)
    assert msg.message == message
    assert msg.severity == severity


def test_JobIOMessage():
    payload = b'Hello world'
    msg = JobIOMessage.create(b'Hello world')
    assert msg.payload == base64.b85encode(payload).decode()
    assert msg.buffer == payload


def test_SessionEndMessage():
    parameters = {"status": JobStatus.PASS.name}
    msg = SessionEndMessage.create(status=JobStatus.PASS)
    assert msg.payload == parameters
    assert msg.status == JobStatus.PASS
    assert msg.job_bucket is None

    # Test the bucket
    bucket = MagicMock()
    bucket.name = "bucketname"
    bucket.access_url.return_value = "http://myurl"

    msg = SessionEndMessage.create(status=JobStatus.PASS,
                                   job_bucket=bucket)
    assert msg.job_bucket.minio_access_url == "http://myurl"
    assert msg.job_bucket.bucket_name == "bucketname"
    bucket.access_url.create_owner_credentials("client")
    bucket.access_url.assert_called_with("client")


def test_Message_send_receive():
    end_session_msg = {
        "status": JobStatus.PASS.name
    }
    expected_header = b'\x00\x00\x00\x59'
    expected_payload = b'{"msg_type": "session_end", "date": "1970-01-01T00:00:00", "payload": {"status": "PASS"}}'

    # Test the sending of messages
    sock_mock = MagicMock()
    send_msg = SessionEndMessage(date=datetime(year=1970, month=1, day=1), payload=end_session_msg)
    assert send_msg.send(sock_mock) == sock_mock.send.return_value
    sock_mock.send.assert_called_with(expected_header+expected_payload)

    # Test the reception of messages
    def side_effect(view, length):
        if length == len(expected_header):
            data = expected_header
        elif length == len(expected_payload):
            data = expected_payload
        else:
            raise ValueError("Unexpected length")  # pragma: no cover

        view[0:len(data)] = data
        return len(data)

    sock_mock = MagicMock(recv_into=MagicMock(side_effect=side_effect))
    recv_msg = Message.next_message(sock_mock)
    assert recv_msg == send_msg


def test_Message_recv_helper():
    def side_effect(view, length):
        if length == 10:
            view[0:6] = b'012345'
            return 6
        elif length == 4:
            view[0:4] = b'6789'
            return 4
        else:
            raise ValueError("Unexpected length")  # pragma: no cover

    sock_mock = MagicMock(recv_into=MagicMock(side_effect=side_effect))
    assert Message.recv(sock_mock, 10) == b'0123456789'

    sock_mock = MagicMock(recv_into=MagicMock(side_effect=[0]))
    with pytest.raises(EOFError) as exc:
        Message.recv(sock_mock, 42)
    assert "The connection got interrupted before receiving the end of the message" in str(exc.value)



from threading import Thread, Event
from dataclasses import asdict, field
from datetime import datetime, timedelta
from ipaddress import IPv4Address
from pathlib import Path

from pydantic.dataclasses import dataclass
from pydantic import validator, PositiveInt
from inotify_simple import INotify, flags
import traceback
import time
import yaml

from .boots import split_mac_addr
from .logger import logger
from .pdu import PDU
from .executor import Executor
from . import config
from . import gitlab


@dataclass
class ConfigPDU:
    driver: str
    config: dict


@dataclass
class ConfigGitlabRunner:
    token: str = "<invalid default>"
    exposed: bool = True

    def verify_or_renew_token(self, gl, description, tags):
        # Remove the runner when not exposed
        if not self.exposed:
            self.remove(gl)
            return

        log_prefix = f"{description}'s {gl.name} runner"

        logger.debug(f"{log_prefix}: Verifying the token {self.token}")
        if not gitlab.verify_runner_token(gitlab_url=gl.url,
                                          token=self.token):
            logger.warning(f"{log_prefix}: The token {self.token} is invalid. "
                           "Starting the renewal process...")

            self.remove(gl)

            runner = gitlab.register_runner(gitlab_url=gl.url,
                                            registration_token=gl.registration_token,
                                            description=description,
                                            tag_list=tags,
                                            maximum_timeout=gl.maximum_timeout)
            if runner:
                self.token = runner.token
                logger.info(f"{log_prefix}: Got assigned the token {self.token}")
            else:
                logger.error(f"{log_prefix}: Could not register the runner on {gl.name}")

    def remove(self, gl):
        if self.token != "<invalid default>":
            print(f"Unregister {gl.name}'s runner {self.token}")

        gitlab.unregister_runner(gitlab_url=gl.url, token=self.token)
        self.token = "<invalid default>"


@dataclass
class ConfigDUT:
    base_name: str
    tags: list[str]
    ip_address: str  # TODO: Make sure all machines have a unique IP
    local_tty_device: str = None
    gitlab: dict[str, ConfigGitlabRunner] = field(default_factory=dict)
    pdu: str = None
    pdu_port_id: str = None
    pdu_off_delay: float = 30
    ready_for_service: bool = False
    is_retired: bool = False
    first_seen: datetime = field(default_factory=lambda: datetime.now())

    @validator('ip_address')
    def ip_address_is_valid(cls, v):
        IPv4Address(v)
        return str(v)

    @property
    def full_name(self):
        # Get the index of the dut by looking at how many duts with
        # the same base name were registered *before* us.
        idx = 1
        for dut in self.mars_db.duts.values():
            if dut.base_name == self.base_name and dut.first_seen < self.first_seen:
                idx += 1

        return f"{config.FARM_NAME}-{self.base_name}-{idx}"

    @property
    def available(self):
        return self.ready_for_service and not self.is_retired

    def expose_on_forges(self):
        # Make sure every gitlab instance is represented in the DUT's config
        for gl in self.mars_db.gitlab.values():
            if self.gitlab.get(gl.name) is None:
                self.gitlab[gl.name] = ConfigGitlabRunner()

        if self.available:
            for gl in self.mars_db.gitlab.values():
                local_cfg = self.gitlab.get(gl.name)
                if gl.expose_runners:
                    local_cfg.verify_or_renew_token(gl, description=self.full_name, tags=self.tags)
                else:
                    local_cfg.remove(gl)
        else:
            # Un-register every associated runner
            for gl_name, local_cfg in self.gitlab.items():
                if gl := self.mars_db.gitlab.get(gl_name):
                    local_cfg.remove(gl)


@dataclass
class ConfigGitlab:
    url: str
    registration_token: str
    expose_runners: bool = True
    maximum_timeout: PositiveInt = 21600
    gateway_runner: ConfigGitlabRunner = None

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent
    def __post_init_post_parse__(self):
        if self.gateway_runner is not None:
            self.gateway_runner.mars_db = self

    @validator("url")
    def url_is_valid(cls, v):
        assert v.startswith("https://")
        return v

    @property
    def should_expose_gateway_runner(self):
        if not self.expose_runners or self.gateway_runner is None:
            return False

        return self.gateway_runner.exposed


@dataclass
class MarsDB:
    pdus: dict[str, ConfigPDU] = field(default_factory=dict)
    duts: dict[str, ConfigDUT] = field(default_factory=dict)
    gitlab: dict[str, ConfigGitlab] = field(default_factory=dict)

    def reset_taint(self):
        self._disk_state = asdict(self)

    @property
    def is_tainted(self):
        return asdict(self) != self._disk_state

    @validator('duts')
    def mac_addresses_are_understood_by_boots(cls, v):
        for addr in v.keys():
            split_mac_addr(addr)
        return v

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent
    def __post_init_post_parse__(self):
        # Since we do not want to repeat ourselves in the config file, the name
        # of objects is set in the parent dict. However, it is quite useful for
        # objects to know their names and have access to the DB. This function
        # adds it back!
        for name, pdu in self.pdus.items():
            pdu.name = name
            pdu.mars_db = self

        for mac_address, dut in self.duts.items():
            dut.mac_address = mac_address
            dut.mars_db = self

        for name, gitlab_instance in self.gitlab.items():
            gitlab_instance.name = name
            gitlab_instance.mars_db = self

        self.reset_taint()

    @classmethod
    def from_file(cls, file_path):
        with open(file_path, 'r') as f:
            data = yaml.safe_load(f)
            return cls(**data if data else {})

    def save(self, file_path):
        with open(file_path, 'w') as f:
            yaml.dump(asdict(self), f, sort_keys=False)
        self.reset_taint()


# TODO: Finish the Machine -> DUT rename
class Machine:
    def __init__(self, db_dut, boots):
        self.db_dut = db_dut
        self.boots = boots

        self.pdu_port = self._create_pdu_port()

        # Executor associated (temporary)
        self.executor = Executor(self)

        # Make sure the updates are reflected in the runner's state
        self.update_config()

    def remove(self):
        self.executor.stop_event.set()
        self.executor.join()

    # Expose all the fields of the associated ConfigDUT object
    def __getattr__(self, attr):
        return getattr(self.db_dut, attr)

    @property
    def id(self):
        return self.mac_address

    @property
    def ready_for_service(self):
        return self.db_dut.ready_for_service

    @ready_for_service.setter
    def ready_for_service(self, val):
        self.db_dut.ready_for_service = val

        # Make sure the updates are reflected in the runner's state
        self.update_config()

        # Make sure the update gets written to disk.
        # HACK: This will force a reload of the configuration and thus
        # update the gitlab runner config
        self.db_dut.mars_db.save(config.MARS_DB_FILE)

    def _create_pdu_port(self):
        config_pdu = self.mars_db.pdus.get(self.db_dut.pdu)
        if config_pdu is None:
            return None

        if pdu := PDU.create(config_pdu.driver, config_pdu.name, config_pdu.config):
            for port in pdu.ports:
                if str(port.port_id) == str(self.db_dut.pdu_port_id):
                    port.min_off_time = self.db_dut.pdu_off_delay
                    return port

            raise ValueError('Could not find a matching port for %s on %s' % (self.db_dut.pdu_port_id, pdu))

    def update_config(self, db_dut=None):
        if db_dut:
            self.db_dut = db_dut

        self.pdu_port = self._create_pdu_port()

        # Make sure the test machines are assigned a static IP address
        # to facilitate testing services hosted on the test
        # machines themselves.
        self.boots.write_network_config(self.mac_address,
                                        self.db_dut.ip_address,
                                        self.full_name)


class Mars(Thread):
    def __init__(self, boots):
        super().__init__(name='MarsClient')

        self.boots = boots
        self.mars_db = None

        self._machines = {}

        self.stop_event = Event()

    @property
    def known_machines(self):
        return list(self._machines.values())

    def get_machine_by_id(self, machine_id, raise_if_missing=False):
        machine = self._machines.get(machine_id)
        if machine is None and raise_if_missing:
            raise ValueError(f"Unknown machine ID '{machine_id}'")
        return machine

    def _machine_update_or_create(self, db_dut):
        machine = self._machines.get(db_dut.mac_address)
        if machine is None:
            machine = Machine(db_dut, self.boots)
            self._machines[machine.mac_address] = machine
        else:
            machine.update_config(db_dut)

        self.mars_db.duts[machine.mac_address] = db_dut

        return machine

    def save_db_if_needed(self):
        if self.mars_db.is_tainted:
            print("Write-back the MarsDB to disk, after some local changes")
            # TODO: Print a list of the changes, possibly using deepdiff?
            self.mars_db.save(config.MARS_DB_FILE)

    def add_or_update_machine(self, fields: dict):
        mac_address = fields.pop("mac_address")

        if db_dut := self.mars_db.duts.get(mac_address):
            cur_state = asdict(db_dut)
            db_dut = ConfigDUT(**(cur_state | fields))
        else:
            db_dut = ConfigDUT(**fields)

        # TODO: Try to find a way not to have to add these fields
        db_dut.mac_address = mac_address
        db_dut.mars_db = self.mars_db

        machine = self._machine_update_or_create(db_dut)
        self.save_db_if_needed()

        return machine

    def sync_machines(self):
        self.mars_db = MarsDB.from_file(config.MARS_DB_FILE)

        local_only_machines = set(self.known_machines)
        for m in self.mars_db.duts.values():
            # Ignore retired machines
            if m.is_retired:
                continue

            machine = self._machine_update_or_create(m)

            # Remove the machine from the list of local-only machines
            local_only_machines.discard(machine)

        # Delete all the machines that are not found in MaRS
        for machine in local_only_machines:
            self._machines[machine.id].remove()
            del self._machines[machine.id]

        # Expose the gateway runners
        for gl in self.mars_db.gitlab.values():
            if gl.should_expose_gateway_runner:
                gl.gateway_runner.verify_or_renew_token(gl,
                                                        description=f"{config.FARM_NAME}-gateway",
                                                        tags=[f"{config.FARM_NAME}-gateway", 'CI-gateway'])
            else:
                gl.gateway_runner.remove(gl)

        # Expose the DUTs on all the forges
        for m in self.mars_db.duts.values():
            m.expose_on_forges()

        # Update the gitlab runner configuration
        gitlab.generate_runner_config(self.mars_db)

        # Save any change that may have happened after reloading
        self.save_db_if_needed()

    def stop(self, wait=True):
        self.stop_event.set()

        # Signal all the executors we want to stop
        for machine in self.known_machines:
            machine.executor.stop_event.set()

        if wait:
            self.join()

    def join(self):
        for machine in self.known_machines:
            machine.executor.join()
        super().join()

    def run(self):
        # Make sure the config file exists
        Path(config.MARS_DB_FILE).touch(exist_ok=True)

        # Set up a watch
        inotify = INotify()
        watch_flags = flags.CREATE | flags.DELETE | flags.MODIFY | flags.DELETE_SELF
        inotify.add_watch(config.MARS_DB_FILE, watch_flags)

        # Now wait for changes to the file
        last_sync = None
        while not self.stop_event.is_set():
            try:
                reason = None
                if last_sync is None:
                    reason = "Initial boot"
                elif len(inotify.read(timeout=1000)) > 0:
                    reason = "Got updated on disk"
                elif datetime.now() - last_sync > timedelta(minutes=30):
                    reason = "Periodic check"

                if reason:
                    logger.info(f"Syncing the MaRS DB. Reason: {reason}")

                    self.sync_machines()
                    last_sync = datetime.now()
            except Exception:
                traceback.print_exc()
                logger.info("Trying again in 60 seconds")
                time.sleep(60)



from unittest.mock import call, patch, MagicMock
from urllib.parse import urlparse
import json

from minio.error import S3Error
import pytest

from server.minioclient import MinioClient, MinIOPolicyStatement, generate_policy
import server.config as config


def test_generate_policy():
    statement1 = MinIOPolicyStatement()
    statement2 = MinIOPolicyStatement(buckets=['bucket1', 'bucket2'],
                                      actions=["action1", "action2"],
                                      source_ips=["ip1", "ip2"])
    statement3 = MinIOPolicyStatement(buckets=['bucket2'],
                                      actions=["action1", "action3"],
                                      allow=False, not_source_ips=["ip3"])

    assert generate_policy([statement1, statement2, statement3]) == {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Action": ["s3:*"],
                "Effect": "Allow",
                "Resource": ['arn:aws:s3:::*', 'arn:aws:s3:::*/*'],
            },
            {
                "Action": ["action1", "action2"],
                "Effect": "Allow",
                "Resource": ['arn:aws:s3:::bucket1', 'arn:aws:s3:::bucket2',
                             'arn:aws:s3:::bucket1/*', 'arn:aws:s3:::bucket2/*'],
                "Condition": {
                    "IpAddress": {
                        "aws:SourceIp": ["ip1", "ip2"]
                    }
                }
            },
            {
                "Action": ["action1", "action3"],
                "Effect": "Deny",
                "Resource": ['arn:aws:s3:::bucket2', 'arn:aws:s3:::bucket2/*'],
                "Condition": {
                    "NotIpAddress": {
                        "aws:SourceIp": ["ip3"]
                    }
                }
            }
        ]
    }


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_client_instantiation__defaults(subproc_mock, minio_mock):
    MinioClient()
    minio_mock.assert_called_once_with(endpoint=urlparse(config.MINIO_URL).netloc,
                                       access_key=config.MINIO_ROOT_USER, secret_key=config.MINIO_ROOT_PASSWORD,
                                       secure=False)
    subproc_mock.assert_called_once_with(['mcli', '-q', '--no-color', 'alias', 'set',
                                          config.MINIO_ADMIN_ALIAS, config.MINIO_URL,
                                          config.MINIO_ROOT_USER, config.MINIO_ROOT_PASSWORD])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_client_instantiation__custom_params(subproc_mock, minio_mock):
    MinioClient(url="http://hello-world", user="accesskey", secret_key="secret_key", alias="toto")
    minio_mock.assert_called_once_with(endpoint="hello-world", access_key="accesskey",
                                       secret_key="secret_key", secure=False)
    subproc_mock.assert_called_once_with(['mcli', '-q', '--no-color', 'alias', 'set',
                                          "toto", "http://hello-world", "accesskey", "secret_key"])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_client_instantiation__no_aliases(subproc_mock, minio_mock):
    MinioClient(url="http://hello-world", user="accesskey", secret_key="secret_key", alias=None)
    minio_mock.assert_called_once_with(endpoint="hello-world", access_key="accesskey",
                                       secret_key="secret_key", secure=False)
    subproc_mock.assert_not_called()


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_client_remove_alias(subproc_mock, minio_mock):
    client = MinioClient(url="http://hello-world", user="accesskey", secret_key="secret_key", alias="toto")
    subproc_mock.assert_called_once_with(['mcli', '-q', '--no-color', 'alias', 'set',
                                          "toto", "http://hello-world", "accesskey", "secret_key"])

    client.remove_alias()
    subproc_mock.assert_called_with(['mcli', '-q', '--no-color', 'alias', 'rm', "toto"])


@patch("subprocess.check_call")
def test_is_local_url(subproc_mock):
    minio = MinioClient(url="http://10.42.0.1:9000")

    assert minio.is_local_url("http://10.42.0.1:9000/toto")
    assert not minio.is_local_url("http://hello-world/toto")

    minio = MinioClient(url="http://hello-world")
    assert not minio.is_local_url("http://10.42.0.1:9000/toto")
    assert minio.is_local_url("http://hello-world/toto")


class MockStream:
    def iter_content(self, _):
        yield b'hello world'

    def raise_for_status(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


@patch("server.minioclient.requests.get", return_value=MockStream())
@patch("server.minioclient.Minio", autospec=True)
@patch("server.minioclient.tempfile.NamedTemporaryFile", autospec=True)
@patch("subprocess.check_call")
def test_save_boot_artifact(subproc_mock, named_temp_mock, minio_mock, get_mock):
    client = MinioClient()

    named_temp_mock().__enter__().name = "/tmp/temp_file"
    client.save_boot_artifact("https://toto.com/path", "/toto/path")

    client._client.fput_object.assert_called_once_with("boot", "/toto/path", "/tmp/temp_file")


@patch("server.minioclient.Minio", autospec=True)
@patch("server.minioclient.TarFile.open", autospec=True)
@patch("subprocess.check_call")
def test_extract_archive(subproc_mock, tarfile_mock, minio_mock):
    client = MinioClient()

    archive_mock = tarfile_mock.return_value.__enter__.return_value
    file_obj = MagicMock()
    members = [MagicMock(isfile=MagicMock(return_value=True), size=42),
               MagicMock(isfile=MagicMock(return_value=False)),
               None]
    members[0].name = "toto"
    archive_mock.next = MagicMock(side_effect=members)

    client.extract_archive(file_obj, "bucket/rootpath")

    tarfile_mock.assert_called_once_with(fileobj=file_obj, mode='r')
    archive_mock.extractfile.assert_called_once_with(members[0])

    client._client.put_object.assert_called_once_with("bucket/rootpath",
                                                      "toto",
                                                      archive_mock.extractfile(),
                                                      members[0].size,
                                                      num_parallel_uploads=1)


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_make_bucket(subproc_mock, minio_mock):
    client = MinioClient()
    client._client = MagicMock()
    client.make_bucket('test-id')
    client._client.make_bucket.assert_called_once_with('test-id')

    def side_effect(*arg, **kwargs):
        raise S3Error('code', 'message', 'resource', 'request_id', 'host_id', 'response')
    client._client.make_bucket.side_effect = side_effect

    with pytest.raises(ValueError) as exc:
        client.make_bucket('test-id')
    assert "The bucket already exists" in str(exc.value)


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_remove_bucket(subproc_mock, minio_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.remove_bucket('test-id')
    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'rb', '--force', 'local/test-id'])])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_minio_add_user(subproc_mock, minio_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.add_user('job-id-c', 'job-password')
    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'admin', 'user', 'add', 'local', 'job-id-c', 'job-password']),
    ])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_minio_remove_user(subproc_mock, minio_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.remove_user('username')
    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'admin', 'user', 'remove', 'local', 'username']),
    ])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
@patch("subprocess.check_output", return_value="""{"status": "success", "accessKey": "username",
    "userStatus": "enabled", "memberOf": ["group1", "group2"]}""")
def test_minio_groups_user_is_in(subproc_mock, _, minio_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    assert client.groups_user_is_in() == ["group1", "group2"]
    assert client.groups_user_is_in('username') == ["group1", "group2"]
    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', '--json', 'admin', 'user', 'info', 'local', 'test']),
        call(['mcli', '-q', '--no-color', '--json', 'admin', 'user', 'info', 'local', 'username'])
    ])


@patch("server.minioclient.Minio", autospec=True)
@patch("subprocess.check_call")
def test_minio_add_user_to_group(subproc_mock, minio_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.add_user_to_group('username', 'groupname')
    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'admin', 'group', 'add', 'local', 'groupname', 'username']),
    ])


@patch("subprocess.check_call")
@patch("server.minioclient.tempfile.NamedTemporaryFile", autospec=True)
def test_minio_add_user_policy_add(named_temp_mock, subproc_mock):
    temp_mock = MagicMock()
    temp_mock.name = '/tmp/temp_file'
    named_temp_mock.return_value.__enter__.return_value = temp_mock

    policy_statements = [MinIOPolicyStatement(['bucket'])]
    expected_policy = generate_policy(policy_statements)

    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.apply_user_policy('policy_name', 'username', policy_statements=policy_statements)

    temp_mock.write.assert_called_once_with(json.dumps(expected_policy).encode())

    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'admin', 'policy', 'add', 'local', 'policy_name', '/tmp/temp_file']),
        call(['mcli', '-q', '--no-color', 'admin', 'policy', 'set', 'local', 'policy_name', 'user=username'])])


@patch("subprocess.check_call")
def test_minio_remove_user_policy(subproc_mock):
    client = MinioClient(url='http://test.invalid', user='test', secret_key='test', alias="local")
    client.remove_user_policy('policy_name', 'username')

    subproc_mock.assert_has_calls([
        call(['mcli', '-q', '--no-color', 'alias', 'set', 'local', 'http://test.invalid', 'test', 'test']),
        call(['mcli', '-q', '--no-color', 'admin', 'policy', 'unset', 'local', 'policy_name', 'user=username']),
        call(['mcli', '-q', '--no-color', 'admin', 'policy', 'remove', 'local', 'policy_name'])])


def test_create_valid_bucket_name():
    # Name is too short
    assert MinioClient.create_valid_bucket_name("") == "b--x"
    assert MinioClient.create_valid_bucket_name("ab") == "b--ab"

    # Name is too long
    bucket_name = "rhjfklsahjfkdlsahuifeohwuiafohuiofhueioahufieohauiefohuaieofhuiffdsarewfgdsa"
    assert len(MinioClient.create_valid_bucket_name(bucket_name)) == 63

    # Wrong characters
    assert MinioClient.create_valid_bucket_name("/*_~!@#$%^&*()_+|") == "x-----------------x"

    # IP address
    assert MinioClient.create_valid_bucket_name("192.168.5.4") == "ip-192.168.5.4"



from enum import Enum
from datetime import datetime, timedelta
from marshmallow import Schema, fields, post_load
from marshmallow.exceptions import ValidationError
from jinja2 import Template
import yaml
import re

from . import config


class Target:
    class Schema(Schema):
        id = fields.Str()
        tags = fields.List(fields.Str())

        @post_load
        def make(self, data, **kwargs):
            if 'id' not in data and 'tags' not in data:
                raise ValueError("The target is neither identified by tags or id. "
                                 "Use empty tags to mean 'any machines'.")

            return Target(**data)

    def __init__(self, id: str = None, tags: list[str] = None):
        self.id = id
        self.tags = tags if tags is not None else []

    def __str__(self):
        return f"<Target: id={self.id}, tags={self.tags}>"

    @classmethod
    def from_job(cls, data):

        schema = cls.Schema()
        return schema.load(data)


class Timeout:
    class Schema(Schema):
        days = fields.Float(missing=0)
        hours = fields.Float(missing=0)
        minutes = fields.Float(missing=0)
        seconds = fields.Float(missing=0)
        milliseconds = fields.Float(missing=0)
        retries = fields.Int(strict=True, missing=0)

        @post_load
        def make(self, data, **kwargs):
            timeout = timedelta(days=data.get("days", 0),
                                hours=data.get("hours", 0),
                                minutes=data.get("minutes", 0),
                                seconds=data.get("seconds", 0),
                                milliseconds=data.get("milliseconds", 0))
            return Timeout(self.context.get('name'), timeout=timeout, retries=data.get('retries'))

    def __init__(self, name: str, timeout: timedelta, retries: int) -> None:
        self.name = name
        self.timeout = timeout
        self.retries = retries

        self.started_at = None
        self.retried = 0

    @property
    def active_for(self):
        if self.started_at is not None:
            return datetime.now() - self.started_at
        else:
            return None

    @property
    def has_expired(self):
        active_for = self.active_for
        return active_for is not None and active_for > self.timeout

    def start(self):
        self.started_at = datetime.now()

    def reset(self, when=None):
        if when is None:
            when = datetime.now()
        self.started_at = when

    def retry(self):
        self.stop()
        self.retried += 1

        return self.retried <= self.retries

    def stop(self):
        self.started_at = None

    def __str__(self):
        return f"<Timeout {self.name}: value={self.timeout}, retries={self.retried}/{self.retries}>"

    @classmethod
    def from_job(cls, name, data):
        schema = Timeout.Schema(context={"name": name})
        return schema.load(data)


class Timeouts:
    class Type(Enum):
        OVERALL = "overall"
        INFRA_SETUP = "infra_setup"
        INFRA_TEARDOWN = "infra_teardown"
        BOOT_CYCLE = "boot_cycle"
        CONSOLE = "console_activity"
        FIRST_CONSOLE_MSG = "first_console_activity"

    class Schema(Schema):
        overall = fields.Nested(Timeout.Schema(context={"name": "overall"}))
        infra_setup = fields.Nested(Timeout.Schema(context={"name": "infra_setup"}))
        infra_teardown = fields.Nested(Timeout.Schema(context={"name": "infra_teardown"}))
        boot_cycle = fields.Nested(Timeout.Schema(context={"name": "boot_cycle"}))
        console_activity = fields.Nested(Timeout.Schema(context={"name": "console_activity"}))
        first_console_activity = fields.Nested(Timeout.Schema(context={"name": "first_console_activity"}))

        @post_load
        def make(self, data, **kwargs):
            timeouts = dict(self.context.get('default_timeouts'))
            for t_type, t_data in data.items():
                timeouts[t_type] = t_data
            return Timeouts(timeouts)

    def __init__(self, timeouts):
        for t_type in Timeouts.Type:
            timeout = timeouts.get(t_type.value)
            if timeout is None:
                timeout = Timeout(name=t_type.value, timeout=timedelta.max, retries=0)

            # Sanity check the timeout
            if t_type in [self.Type.OVERALL, self.Type.INFRA_TEARDOWN] and timeout.retries != 0:
                raise ValueError("Neither the overall nor the teardown timeout can have retries")

            setattr(self, t_type.value, timeout)

    def __iter__(self):
        for t_type in Timeouts.Type:
            yield getattr(self, t_type.value)

    @property
    def expired_list(self):
        expired = []
        for timeout in self:
            if timeout.has_expired:
                expired.append(timeout)
        return expired

    @property
    def has_expired(self):
        return len(self.expired_list) > 0

    @classmethod
    def from_job(cls, data, defaults=None):
        schema = cls.Schema(context={"default_timeouts": defaults if defaults is not None else {}})
        return schema.load(data)


class ConsoleState:
    class Schema(Schema):
        class PatternSchema(Schema):
            regex = fields.Str()

        session_end = fields.Nested(PatternSchema(), missing=None)
        session_reboot = fields.Nested(PatternSchema(), missing=None)
        job_success = fields.Nested(PatternSchema(), missing=None)
        job_warn = fields.Nested(PatternSchema(), missing=None)

        @post_load
        def make(self, data, **kwargs):
            patterns = {
                "session_end": "^\\[[\\d \\.]{12}\\] reboot: Power Down$",
                "session_reboot": None,
                "job_success": None,
                "job_warn": None
            }

            for name, pattern in data.items():
                if pattern is not None:
                    if regex := pattern.get('regex', None):
                        patterns[name] = regex

            return ConsoleState(**patterns)

    def __init__(self, session_end, session_reboot, job_success, job_warn):
        self.session_end = session_end
        self.session_reboot = session_reboot
        self.job_success = job_success
        self.job_warn = job_warn

        self._regexs = {
            "session_end": re.compile(session_end.encode()),
        }

        if session_reboot is not None:
            self._regexs["session_reboot"] = re.compile(session_reboot.encode())

        if job_success is not None:
            self._regexs["job_success"] = re.compile(job_success.encode())

        if job_warn is not None:
            self._regexs["job_warn"] = re.compile(job_warn.encode())

        self._matched = set()

    def process_line(self, line):
        matched = set()
        for name, regex in self._regexs.items():
            if regex.search(line):
                matched.add(name)

        # Extend the list of matched regex
        self._matched.update(matched)

        return matched

    def reset_per_boot_state(self):
        self._matched.discard("session_reboot")

    @property
    def session_has_ended(self):
        return "session_end" in self._matched

    @property
    def needs_reboot(self):
        return "session_reboot" in self._matched

    @property
    def job_status(self):
        if "session_end" not in self._matched:
            return "INCOMPLETE"

        if "job_success" in self._regexs:
            if "job_success" in self._matched:
                if "job_warn" in self._matched:
                    return "WARN"
                else:
                    return "PASS"
            else:
                return "FAIL"
        else:
            return "COMPLETE"

    @classmethod
    def from_job(cls, data):
        schema = cls.Schema()
        return schema.load(data)


def _multiline_string(lines):
    if isinstance(lines, str):
        return lines
    elif isinstance(lines, list):
        return " ".join(lines)
    else:
        assert False


class Deployment:
    def __init__(self, kernel_url=None, initramfs_url=None, kernel_cmdline=None):
        self.kernel_url = kernel_url
        self.kernel_cmdline = kernel_url
        self.initramfs_url = initramfs_url

    def update(self, data):
        if (kernel_url := data.get('kernel', {}).get('url')) is not None:
            self.kernel_url = kernel_url

        if (kernel_cmdline := data.get('kernel', {}).get('cmdline')) is not None:
            self.kernel_cmdline = kernel_cmdline

        if (initramfs_url := data.get('initramfs', {}).get('url')) is not None:
            self.initramfs_url = initramfs_url

    def __str__(self):
        return f"""<Deployment:
    kernel_url: {self.kernel_url}
    initramfs_url: {self.initramfs_url}
    kernel_cmdline: {self.kernel_cmdline}>
"""


class Job:
    class Schema(Schema):
        class DeploymentsSchema(Schema):
            class DeploymentSchema(Schema):
                class KernelSchema(Schema):
                    url = fields.Str()
                    cmdline = fields.Method("get_cmdline", deserialize="load_cmdline")

                    def get_cmdline(self, obj):  # pragma: nocover
                        return obj.kernel_cmdline

                    def load_cmdline(self, value):
                        return _multiline_string(value)

                class InitramfsSchema(Schema):
                    url = fields.Str()

                kernel = fields.Nested(KernelSchema())
                initramfs = fields.Nested(InitramfsSchema())

            start = fields.Nested(DeploymentSchema(), required=True)
            cont = fields.Nested(DeploymentSchema(), data_key="continue", attribute="continue")

        version = fields.Int(strict=True, missing=1)
        deadline = fields.DateTime(missing=datetime.max)
        target = fields.Nested(Target.Schema(), missing=Target())
        timeouts = fields.Nested(Timeouts.Schema(), missing=None)
        console_patterns = fields.Nested(ConsoleState.Schema(), required=True)
        deployment = fields.Nested(DeploymentsSchema(), required=True)

        @post_load
        def make(self, data, **kwargs):
            # Timeouts
            if data.get('timeouts') is None:
                data['timeouts'] = Timeouts(self.context.get('default_timeouts'))

            # Deployments
            deployment = data.pop('deployment', {})

            deployment_start = Deployment()
            deployment_start.update(deployment['start'])
            data['deployment_start'] = deployment_start

            # Source the default 'continue' deployment from the start one, then
            # update with the continue one.
            deployment_continue = Deployment()
            deployment_continue.update(deployment['start'])
            deployment_continue.update(deployment.get('continue', {}))
            data['deployment_continue'] = deployment_continue

            data['bucket'] = self.context.get('bucket')

            return Job(**data)

    def __init__(self, version, deadline, target, timeouts, console_patterns, deployment_start,
                 deployment_continue, bucket=None):
        self.version = version
        self.deadline = deadline
        self.target = target
        self.timeouts = timeouts
        self.console_patterns = console_patterns
        self.deployment_start = deployment_start
        self.deployment_continue = deployment_continue

    @classmethod
    def from_job(cls, job_yml, bucket=None):
        j = yaml.safe_load(job_yml)

        default_timeouts = {
            "overall": Timeout(name="overall", timeout=timedelta(hours=6), retries=0),
            "infra_teardown": Timeout(name="infra_teardown", timeout=timedelta(minutes=10), retries=0),
        }

        try:
            schema = cls.Schema(context={
                'default_timeouts': default_timeouts,
                'bucket': bucket,
            })
            return schema.load(j)
        except ValidationError as e:
            raise ValueError(str(e))

    @classmethod
    def render_with_resources(cls, job_str, machine, bucket=None):
        template_params = {
            "ready_for_service": machine.ready_for_service,
            "machine_id": machine.id,
            "machine_tags": machine.tags,
            "local_tty_device": machine.local_tty_device,
            **{k.lower(): v for k, v in config.job_environment_vars().items()},
        }

        if bucket:
            dut_creds = bucket.credentials('dut')

            template_params["minio_url"] = bucket.minio.url
            template_params["job_bucket"] = bucket.name
            template_params["job_bucket_access_key"] = dut_creds.username
            template_params["job_bucket_secret_key"] = dut_creds.password

        rendered_job_str = Template(job_str).render(**template_params)
        return cls.from_job(rendered_job_str)

    @classmethod
    def from_path(cls, job_template_path, machine, bucket=None):
        with open(job_template_path, "r") as f_template:
            template_str = f_template.read()
            return Job.render_with_resources(template_str, machine, bucket)

    def __str__(self):
        return f"""<Job:
    version: {self.version}

    deadline: {self.deadline}
    target: {self.target}

    timeouts:
        overall:                {self.timeouts.overall}
        infra_setup:            {self.timeouts.infra_setup}
        boot_cycle:             {self.timeouts.boot_cycle}
        console_activity:       {self.timeouts.console_activity}
        first_console_activity: {self.timeouts.first_console_activity}

    console patterns:
        session_end:    {self.console_patterns.session_end}
        session_reboot: {self.console_patterns.session_reboot}
        job_success:    {self.console_patterns.job_success}
        job_warn:       {self.console_patterns.job_warn}

    start deployment:
        kernel_url:     {self.deployment_start.kernel_url}
        initramfs_url:  {self.deployment_start.initramfs_url}
        kernel_cmdline: {self.deployment_start.kernel_cmdline}

    continue deployment:
        kernel_url:     {self.deployment_continue.kernel_url}
        initramfs_url:  {self.deployment_continue.initramfs_url}
        kernel_cmdline: {self.deployment_continue.kernel_cmdline}>"""



from logging import getLogger, getLevelName, Formatter, StreamHandler


logger = getLogger(__name__)
logger.setLevel(getLevelName('DEBUG'))
log_formatter = \
    Formatter("%(asctime)s [%(threadName)s] [%(levelname)s] %(funcName)s: "
              "%(message)s")
console_handler = StreamHandler()
console_handler.setFormatter(log_formatter)
logger.addHandler(console_handler)



from unittest.mock import MagicMock, patch, mock_open

from server.gitlab import GitlabRunnerRegistration, register_runner, unregister_runner
from server.gitlab import verify_runner_token, generate_runner_config
import server.config as config


def test_register_runner():
    url = "my url"
    registration_token = "reg_token"
    description = "toto"
    tag_list = ["tag1", "tag2"]

    runner_id = 1234
    runner_token = "my token"

    post_mock_return_value = MagicMock(status_code=201,
                                       json=MagicMock(return_value={"id": runner_id, "token": runner_token}))
    with patch("server.gitlab.requests.post", return_value=post_mock_return_value) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list)
        assert r == GitlabRunnerRegistration(id=runner_id, token=runner_token)

        post_mock.assert_called_with(f"{url}/api/v4/runners",
                                     params={'token': registration_token,
                                             'description': description,
                                             'tag_list': ",".join(tag_list),
                                             'run_untagged': False,
                                             'maximum_timeout': 3600})

    with patch("server.gitlab.requests.post", return_value=MagicMock(status_code=403)) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list)
        assert r is None


def test_unregister_runner():
    url = "my url"
    runner_token = "my token"

    with patch("server.gitlab.requests.delete") as delete_mock:
        unregister_runner(gitlab_url=url, token=runner_token)
        delete_mock.assert_called_with(f"{url}/api/v4/runners", params={"token": runner_token})


def test_verify_runner_token():
    url = "my url"
    runner_token = "my token"

    with patch("server.gitlab.requests.post", return_value=MagicMock(status_code=403)) as post_mock:
        verify_runner_token(gitlab_url=url, token=runner_token)
        post_mock.assert_called_with(f"{url}/api/v4/runners/verify", params={"token": runner_token})


def test_generate_runner_config():
    template_data = "data"

    with patch("server.gitlab.Template") as template_mock:
        with patch("builtins.open", mock_open(read_data=template_data)) as mock_file:
            mars_db = MagicMock()

            generate_runner_config(mars_db)

            template_mock.assert_called_with(template_data)
            template_mock.return_value.render.assert_called_with(config=config, mars_db=mars_db)

            mock_file.return_value.write.assert_called_with(template_mock.return_value.render.return_value)



from easysnmp import snmp_get, snmp_set, snmp_walk
from .. import logger, PDU, PDUPort, PDUState
from typing import Dict

import random
import time


def _is_int(s):
    try:
        s = int(s)
        return True
    except ValueError:
        return False


def retry_on_known_errors(func):
    retriable_errors = [
        "<built-in function set> returned NULL without setting an error",
        "<built-in function get> returned NULL without setting an error",
        "<built-in function walk> returned NULL without setting an error",
    ]

    def retry(*args, **kwargs):
        retries = 3

        for i in range(retries):
            try:
                return func(*args, **kwargs)
            except SystemError as e:
                if str(e) in retriable_errors:
                    logger.warning(f"Caught the re-triable error '{str(e)}', retrying ({i+1}/{retries})")

                    # Wait a minimum of 1 second, plus a random delay to reduce the chances of concurrent requests
                    time.sleep(1 + random.random())
                    continue
                raise e

        raise ValueError(f"The function {func} failed {retries} times in a row")

    return retry


class SnmpPDU(PDU):
    oid_enterprise = '1.3.6.1.4.1'

    def __init__(self, name, config):
        self.hostname = config.get('hostname', None)
        if not self.hostname:
            raise ValueError('SnmpPDU requires a "hostname" configuration key')
        self.community = config.get('community', 'private')

        assert self.system_id
        assert self.outlet_labels
        assert self.outlet_status

        if not hasattr(self, 'outlet_ctrl'):
            # Some PDUs offer a RW status tree, others require a separate
            # tree for writes. Default to the seemingly more common case
            # of a RW tree.
            self.outlet_ctrl = self.outlet_status

        # FIXME: THe UNKNOWN status is a bit of an odd one, not all PDUs expose such a concept.
        assert self.state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])
        if not hasattr(self, 'inverse_state_mapping'):
            self.inverse_state_mapping: Dict[int, PDUState] = \
                dict([(value, key) for key, value in self.state_mapping.items()])
        else:
            assert self.inverse_state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])
        super().__init__(name)

    @property
    def outlet_system_id(self):
        return f'{self.oid_enterprise}.{self.system_id}'

    @property
    def outlet_labels_oid(self):
        return f'{self.outlet_system_id}.{self.outlet_labels}'

    def outlet_status_oid(self, port_id):
        assert isinstance(port_id, int)
        return f'{self.outlet_system_id}.{self.outlet_status}.{port_id}'

    def outlet_ctrl_oid(self, port_id: int):
        assert isinstance(port_id, int)
        return f'{self.outlet_system_id}.{self.outlet_ctrl}.{port_id}'

    @property
    def ports(self):
        ports = []

        try:
            names = [x.value for x in
                     snmp_walk(self.outlet_labels_oid,
                               hostname=self.hostname,
                               community=self.community,
                               version=1)]
        except SystemError as e:
            raise ValueError(f"The snmp_walk() call failed with the following error: {e}")

        for i, name in enumerate(names):
            ports.append(PDUPort(self, i+1, name))

        return ports

    def _port_spec_to_int(self, port_spec):
        if _is_int(port_spec):
            return port_spec
        else:
            for port in self.ports:
                if port.label == port_spec:
                    return port.port_id
            raise ValueError(
                f"{port_spec} can not be interpreted as a valid port")

    @retry_on_known_errors
    def set_port_state(self, port_spec, state):
        SNMP_INTEGER_TYPE = 'i'

        port_id = self._port_spec_to_int(port_spec)
        logger.debug('setting OID %s to state %s with value %d',
                     self.outlet_ctrl_oid(port_id),
                     state,
                     self.state_mapping[state])
        ret = snmp_set(self.outlet_ctrl_oid(port_id),
                       self.state_mapping[state],
                       SNMP_INTEGER_TYPE,
                       hostname=self.hostname,
                       version=1,
                       community=self.community)

        if self.state_transition_delay_seconds is not None:
            logger.debug("Enforcing %s seconds of delay for state change", self.state_transition_delay_seconds)
            # TODO: keep track of state changes to avoid a forced sleep.
            # TODO: Polling for the state change would be better in general.
            # The root cause of this is because PDUs maintain their
            # own configurables how long to delay between
            # transitions, we should probably control that via SNMP,
            # as well.
            time.sleep(self.state_transition_delay_seconds)

        return ret

    @retry_on_known_errors
    def get_port_state(self, port_spec):
        port_id = self._port_spec_to_int(port_spec)
        vs = snmp_get(self.outlet_status_oid(port_id),
                      hostname=self.hostname,
                      version=1,
                      community=self.community)
        value = int(vs.value)
        logger.debug('retrieved OID %s with value %d, maps to state %s',
                     self.outlet_status_oid(port_id),
                     value,
                     self.inverse_state_mapping[value])
        return self.inverse_state_mapping[value]

    def __eq__(self, other):
        return not any([
            getattr(self, attr, None) != getattr(other, attr, None)
            for attr in ["name",
                         "hostname",
                         "community",
                         "system_id",
                         "outlet_labels",
                         "outlet_status",
                         "outlet_ctrl",
                         "state_mapping",
                         "inverse_state_mapping"]])


class ManualSnmpPDU(SnmpPDU):
    def __init__(self, name, config):
        self.system_id = config.get('system_id')
        self.outlet_labels = config.get('outlet_labels')
        self.outlet_status = config.get('outlet_status')
        # Some PDUs offer a RW status tree, others require a separate
        # tree for writes. Default to the seemingly more common case
        # of a RW tree.
        self.outlet_ctrl = config.get('outlet_ctrl', self.outlet_status)

        def populate_state_mapping(state_mapping, d):
            for state, internal_value in d.items():
                v = int(internal_value)
                if state.lower() == "on":
                    state_mapping[PDUState.ON] = v
                elif state.lower() == "off":
                    state_mapping[PDUState.OFF] = v
                elif state.lower() == "reboot":
                    state_mapping[PDUState.REBOOT] = v
                    # Unknown deliberately excluded.
        self.state_mapping = {}
        populate_state_mapping(self.state_mapping,
                               config.get('state_mapping', {}))
        if 'inverse_state_mapping' in config:
            self.inverse_state_mapping = {}
            populate_state_mapping(self.inverse_state_mapping,
                                   config.get('inverse_state_mapping'))

        super().__init__(name, config)



from urllib.parse import urlparse
from dataclasses import dataclass, field
from collections import defaultdict
from tarfile import TarFile
from minio import Minio
from minio.helpers import check_bucket_name
from minio.error import S3Error
from typing import List

import subprocess
import ipaddress
import tempfile
import requests
import json
import re

from . import config


@dataclass
class MinIOPolicyStatement:
    # NOTE: Using the default factory to avoid mutable defaults
    buckets: List[str] = field(default_factory=lambda: ["*"])
    actions: List[str] = field(default_factory=lambda: ["s3:*"])
    allow: bool = True

    # Conditions
    source_ips: List[str] = None
    not_source_ips: List[str] = None


def generate_policy(statements):
    def nesteddict():
        return defaultdict(nesteddict)

    rendered_statements = []
    for s in statements:
        resources = [f"arn:aws:s3:::{b}" for b in s.buckets]
        resources.extend([f"arn:aws:s3:::{b}/*" for b in s.buckets])

        statement = {
            'Action': s.actions,
            'Effect': "Allow" if s.allow else "Deny",
            'Resource': resources,
        }

        conditions = nesteddict()
        if s.source_ips and len(s.source_ips) > 0:
            conditions["IpAddress"]["aws:SourceIp"] = s.source_ips
        if s.not_source_ips and len(s.not_source_ips) > 0:
            conditions["NotIpAddress"]["aws:SourceIp"] = s.not_source_ips
        if len(conditions) > 0:
            statement["Condition"] = conditions

        rendered_statements.append(statement)

    return {
        "Version": "2012-10-17",
        "Statement": rendered_statements
    }


# TODO: rename the methods to be on the form $object_$operation
# to make auto-completion work more efficiently.
class MinioClient():
    def __init__(self,
                 url=config.MINIO_URL,
                 user=config.MINIO_ROOT_USER,
                 secret_key=config.MINIO_ROOT_PASSWORD,
                 alias=config.MINIO_ADMIN_ALIAS):
        self.url = url
        self.user = user
        self.secret_key = secret_key
        self.alias = alias

        self._client = Minio(
            endpoint=urlparse(url).netloc,
            access_key=user,
            secret_key=secret_key,
            secure=False,
        )

        # Some operations can only be used using the commandline tool,
        # so initialize it
        if alias is not None:
            try:
                subprocess.check_call(
                    ["mcli", "-q", "--no-color", "alias", "set", self.alias, url,
                     self.user, self.secret_key])
            except subprocess.CalledProcessError:  # pragma: nocover
                raise ValueError("Invalid credentials") from None

    def remove_alias(self):
        if self.alias is not None:
            subprocess.check_call(["mcli", "-q", "--no-color", "alias", "rm", self.alias])

    def is_local_url(self, url):
        return url.startswith(f"{self.url}/")

    def save_boot_artifact(self, remote_artifact_url, minio_object_name):
        minio_bucket_name = 'boot'
        with tempfile.NamedTemporaryFile("wb") as temp_download_area, \
             requests.get(remote_artifact_url, stream=True) as r:
            r.raise_for_status()
            # Read all the available data, then write to disk
            for chunk in r.iter_content(None):
                temp_download_area.write(chunk)
            temp_download_area.flush()
            self._client.fput_object(minio_bucket_name, minio_object_name,
                                     temp_download_area.name)

    def extract_archive(self, archive_fileobj, bucket_name):
        with TarFile.open(fileobj=archive_fileobj, mode='r') as archive:
            while (member := archive.next()) is not None:
                # Ignore everything that isn't a file
                if not member.isfile():
                    continue
                self._client.put_object(bucket_name, member.name, archive.extractfile(member),
                                        member.size, num_parallel_uploads=1)

    def make_bucket(self, bucket_name):
        try:
            self._client.make_bucket(bucket_name)
        except S3Error:
            raise ValueError("The bucket already exists") from None

    # NOTE: Using minioclient's remove_bucket requires first to empty the
    # bucket. Use the CLI version for now.
    def remove_bucket(self, bucket_name):
        assert self.alias is not None

        subprocess.check_call(["mcli", "-q", "--no-color", "rb", "--force",
                               f'{self.alias}/{bucket_name}'])

    def add_user(self, user_id, password):
        assert self.alias is not None

        subprocess.check_call(["mcli", "-q", "--no-color", "admin", "user", "add",
                               self.alias, user_id, password])

    def remove_user(self, user_id):
        assert self.alias is not None

        subprocess.check_call(["mcli", "-q", "--no-color", "admin", "user", "remove",
                               self.alias, user_id])

    def groups_user_is_in(self, user_id=None):
        assert self.alias is not None

        if user_id is None:
            user_id = self.user

        output = subprocess.check_output(["mcli", "-q", "--no-color", "--json", "admin",
                                          "user", "info", self.alias, user_id])
        return json.loads(output).get('memberOf', [])

    def add_user_to_group(self, user_id, group_name):
        assert self.alias is not None

        subprocess.check_call(["mcli", "-q", "--no-color", "admin", "group", "add",
                               self.alias, group_name, user_id])

    def apply_user_policy(self, policy_name, user_id, policy_statements):
        assert self.alias is not None

        with tempfile.NamedTemporaryFile(suffix='json') as f:
            policy = generate_policy(policy_statements)
            f.write(json.dumps(policy).encode())
            f.flush()

            subprocess.check_call(["mcli", "-q", "--no-color", "admin", "policy",
                                   "add", self.alias, policy_name, f.name])
            subprocess.check_call(["mcli", "-q", "--no-color", "admin", "policy", "set",
                                   self.alias, policy_name, f"user={user_id}"])

    def remove_user_policy(self, policy_name, user_id):
        assert self.alias is not None

        subprocess.check_call(["mcli", "-q", "--no-color", "admin", "policy", "unset",
                               self.alias, policy_name, f"user={user_id}"])

        subprocess.check_call(["mcli", "-q", "--no-color", "admin", "policy", "remove",
                               self.alias, policy_name])

    @classmethod
    def create_valid_bucket_name(cls, base_name):
        # Bucket names can consist only of lowercase letters, numbers, dots (.), and hyphens (-)
        name = re.sub(r'[^a-z0-9\-\.]', '-', base_name)

        # Bucket names must be between 3 and 63 characters long.
        if len(name) < 3:
            name = "b--" + name

        # Bucket names must begin and end with a letter or number.
        # Bucket names can't begin with xn-- (for buckets created after February 2020)
        if name.startswith('xn--') or name[0] == '.' or name[0] == '-':
            name = 'x' + name
        if name[-1] == '.' or name[-1] == '-':
            name += 'x'

        # Bucket names must not be formatted as an IP address (for example, 192.168.5.4)
        try:
            ipaddress.ip_address(name)

            # The name is an ip address, add a prefix!
            name = "ip-" + name
        except ValueError:
            # The name isn't an IP address, all is good!
            pass

        # Bucket names must be between 3 and 63 characters long.
        if len(name) >= 63:
            name = name[0:63]

        # Do the final checks
        check_bucket_name(name)

        return name



from pathlib import Path
from unittest.mock import patch
from server.boots import (
    split_mac_addr,
    parse_dhcp_hosts,
    Dnsmasq,
    BootService,
)
import glob
import os
import pytest
import shutil

import server


def test_split_mac_addr():
    assert split_mac_addr('00:01:02:03:04:05') == \
        ['00', '01', '02', '03', '04', '05']
    assert split_mac_addr('00-01-02-03-04-05') == \
        ['00', '01', '02', '03', '04', '05']
    assert split_mac_addr('000102030405') == \
        ['00', '01', '02', '03', '04', '05']
    with pytest.raises(ValueError):
        # Too long
        split_mac_addr('00:01:02:03:04:05:06')
    with pytest.raises(ValueError):
        # Too short
        split_mac_addr('00:01:02:03:04')
    with pytest.raises(ValueError):
        # Bad separator
        split_mac_addr('00:01:02:03:04?05')
    with pytest.raises(ValueError):
        split_mac_addr('garbage')


def test_parse_dhcp_hosts():
    hosts = parse_dhcp_hosts("""
# A comment is fine
10:62:e5:04:39:89,10.42.0.10,set:igalia-amd-gfx8-1


10:62:e5:0e:0a:54,10.42.0.11,set:igalia-amd-gfx8-2
""")
    assert len(hosts) == 2
    assert ('10:62:e5:04:39:89', '10.42.0.10', 'igalia-amd-gfx8-1') in \
        hosts
    assert ('10:62:e5:0e:0a:54', '10.42.0.11', 'igalia-amd-gfx8-2') in \
        hosts
    hosts = parse_dhcp_hosts("")
    assert len(hosts) == 0
    with pytest.raises(RuntimeError):
        # Deleted first comma
        hosts = parse_dhcp_hosts("""
10:62:e5:04:39:8910.42.0.10,set:igalia-amd-gfx8-1
""")


@patch.object(server.boots.Dnsmasq, '_wait_for_dnsmasq_to_fork')
@patch("subprocess.Popen")
def test_dnsmasq_no_binary(popen_mock, dnsmasq_waiter, tmp_path):
    with patch.object(shutil, 'which', lambda _: False), pytest.raises(RuntimeError):
        _ = Dnsmasq('br0', {})


@patch.object(server.boots.Dnsmasq, '_wait_for_dnsmasq_to_fork')
@patch("subprocess.Popen")
def test_dnsmasq_launch(popen_mock, dnsmasq_waiter, tmp_path):
    paths = {
        'BOOTS_ROOT': tmp_path,
        'TFTP_DIR': f'{tmp_path}/tftp',
        'PXELINUX_CONFIG_DIR': f'{tmp_path}/pxelinux.cfg',
    }
    _ = Dnsmasq('br0', paths)
    dnsmasq_waiter.assert_called_once()
    popen_mock.assert_called_once_with(
        ['dnsmasq',
         '--port=0',
         f'--pid-file={tmp_path}/dnsmasq.pid',
         f'--dhcp-hostsfile={tmp_path}/hosts.dhcp',
         f'--dhcp-optsfile={tmp_path}/options.dhcp',
         f'--dhcp-leasefile={tmp_path}/dnsmasq.leases',
         '--dhcp-match=set:efi-x86_64,option:client-arch,7',
         '--dhcp-boot=tag:efi-x86_64,syslinux.efi',
         '--dhcp-boot=lpxelinux.0',
         '--dhcp-range=10.42.0.10,10.42.0.100',
         '--dhcp-script=/bin/echo',
         '--enable-tftp=br0',
         f'--tftp-root={tmp_path}/tftp',
         f'--log-facility={tmp_path}/dnsmasq.log',
         '--log-queries=extra',
         '--conf-file=/dev/null',
         '--interface=br0'], bufsize=0)


@patch.object(server.boots.Dnsmasq, 'reload')
@patch.object(server.boots.Dnsmasq, '_wait_for_dnsmasq_to_fork')
@patch("subprocess.Popen")
def test_dnsmasq_add_static_address(popen_mock, dnsmasq_waiter,
                                    dnsmasq_reloader, tmp_path):
    paths = {
        'BOOTS_ROOT': tmp_path,
        'TFTP_DIR': f'{tmp_path}/tftp',
        'PXELINUX_CONFIG_DIR': f'{tmp_path}/pxelinux.cfg',
    }
    os.makedirs(paths['PXELINUX_CONFIG_DIR'])
    dnsmasq = Dnsmasq('br0', paths)
    dnsmasq.add_static_address('00:11:22:33:44:55',
                               '1.2.3.4',
                               'hostname')
    dnsmasq.add_static_address('00:11:22:33:44:66',
                               '1.2.3.5',
                               'hostname2')
    p = Path(paths['BOOTS_ROOT']) / "hosts.dhcp"
    assert p.read_text() == \
        """00:11:22:33:44:55,1.2.3.4,set:hostname
00:11:22:33:44:66,1.2.3.5,set:hostname2
"""
    assert len(dnsmasq_reloader.mock_calls) == 2


@patch("server.boots.Dnsmasq", autospec=True)
@patch("server.boots.provision_network_boot_service")
def test_bootservice(mock_provisioner, mock_dnsmasq, tmp_path):
    paths = {
        'BOOTS_ROOT': tmp_path,
        'TFTP_DIR': f'{tmp_path}/tftp',
        'PXELINUX_CONFIG_DIR': f'{tmp_path}/pxelinux.cfg',
    }
    os.makedirs(paths['PXELINUX_CONFIG_DIR'])
    service = BootService(private_interface='br0',
                          default_kernel='default_kernel',
                          default_initrd='default_initrd',
                          default_cmdline='default_cmdline',
                          config_paths=paths)
    mock_provisioner.assert_called_once()
    service.stop()
    service.dnsmasq.stop.assert_called_once()


@patch("server.boots.Dnsmasq", autospec=True)
@patch("server.boots.provision_network_boot_service")
def test_bootservice_write_config(mock_provisioner, mock_dnsmasq, tmp_path):
    paths = {
        'BOOTS_ROOT': tmp_path,
        'TFTP_DIR': f'{tmp_path}/tftp',
        'PXELINUX_CONFIG_DIR': f'{tmp_path}/pxelinux.cfg',
    }
    os.makedirs(paths['PXELINUX_CONFIG_DIR'])
    service = BootService(private_interface='br0',
                          default_kernel='default_kernel',
                          default_initrd='default_initrd',
                          default_cmdline='default_cmdline',
                          config_paths=paths)

    service.write_network_config('00:11:22:33:44:55',
                                 '1.2.3.4',
                                 'hostname')
    service.dnsmasq.add_static_address.assert_called_with(
        '00:11:22:33:44:55',
        '1.2.3.4',
        'hostname')

    service.write_pxelinux_config('default',
                                  'kernel_path',
                                  'cmdline',
                                  'initrd_path')
    p = Path(paths['PXELINUX_CONFIG_DIR']) / "default"
    assert p.read_text() == \
        """DEFAULT default
label default
  KERNEL kernel_path
  INITRD initrd_path
  APPEND cmdline"""

    service.write_pxelinux_config('00:11:22:33:44:55',
                                  'kernel_path',
                                  'cmdline',
                                  'initrd_path')
    p = Path(paths['PXELINUX_CONFIG_DIR']) / "01-00-11-22-33-44-55"
    assert p.read_text() == \
        """DEFAULT default
label default
  KERNEL kernel_path
  INITRD initrd_path
  APPEND cmdline"""

    service.stop()
    config_files = \
        glob.glob(os.path.join(paths['PXELINUX_CONFIG_DIR'], '01-*'))
    assert len(config_files) == 0


@patch("server.boots.Dnsmasq", autospec=True)
@patch("server.boots.provision_network_boot_service")
def test_bootservice_remove_config(mock_provisioner, mock_dnsmasq, tmp_path):
    paths = {
        'BOOTS_ROOT': tmp_path,
        'TFTP_DIR': f'{tmp_path}/tftp',
        'PXELINUX_CONFIG_DIR': f'{tmp_path}/pxelinux.cfg',
    }
    os.makedirs(paths['PXELINUX_CONFIG_DIR'])
    service = BootService(private_interface='br0',
                          default_kernel='default_kernel',
                          default_initrd='default_initrd',
                          default_cmdline='default_cmdline',
                          config_paths=paths)

    service.write_pxelinux_config('00:11:22:33:44:55',
                                  'kernel_path',
                                  'cmdline',
                                  'initrd_path')
    p = Path(paths['PXELINUX_CONFIG_DIR']) / "01-00-11-22-33-44-55"
    assert p.exists()

    service.remove_pxelinux_config('00:11:22:33:44:55')
    assert not p.exists()
